#! /usr/bin/env python
#
# Test(s) for ../text_categorizer.py
#
# Notes:
# - This can be run as follows:
#   $ PYTHONPATH=".:$PYTHONPATH" python ./mezcla/tests/test_<module>.py
#

"""Tests for text_categorizer module"""

# Standard packages
## OLD: import random

# Installed packages
import pytest

# Local packages
from mezcla.unittest_wrapper import TestWrapper, trap_exception
from mezcla import debug
from mezcla import glue_helpers as gh
from mezcla.my_regex import my_re
from mezcla import system
from mezcla import misc_utils

# Note: Two references are used for the module to be tested:
#    THE_MODULE:	    global module object
system.setenv("USE_XGB", "1")
import mezcla.text_categorizer as THE_MODULE

# Environment options
TEST_TBD = system.getenv_bool("TEST_TBD", False,
                              description="Test features to be designed: TBD")


class TestTextCategorizerUtils(TestWrapper):
    """Class for utility test cases"""
    script_module = TestWrapper.get_testing_module_name(__file__, THE_MODULE)

    @pytest.mark.xfail                   # TODO: remove xfail
    def test_sklearn_report(self):
        """Ensure sklearn_report works as expected"""
        debug.trace(4, "test_sklearn_report()")
        assert(False)

    @pytest.mark.xfail                   # TODO: remove xfail
    def test_create_tabular_file(self):
        """Ensure create_tabular_file works as expected"""
        debug.trace(4, "test_create_tabular_file()")
        assert(False)

    @pytest.mark.xfail                   # TODO: remove xfail
    def test_read_categorization_data(self):
        """Ensure read_categorization_data works as expected"""
        debug.trace(4, "test_read_categorization_data()")
        assert(False)

    def test_int_if_whole(self):
        """Ensure int_if_whole works as expected"""
        debug.trace(4, "test_int_if_whole()")
        assert THE_MODULE.int_if_whole(2.0) == 2
        assert THE_MODULE.int_if_whole(2.999) == 2.999

    @pytest.mark.xfail                   # TODO: remove xfail
    def test_param_or_default(self):
        """Ensure param_or_default works as expected"""
        debug.trace(4, "test_param_or_default()")
        assert(False)

    @pytest.mark.xfail                   # TODO: remove xfail
    def test_data_file(self):
        """Makes sure TODO works as expected"""
        debug.trace(4, "TestTextCategorizer.test_data_file()")
        data = ["TODO1", "TODO2"]
        gh.write_lines(self.temp_file, data)
        output = self.run_script("", self.temp_file)
        assert my_re.search(r"TODO-pattern", output.strip())
        return

    @pytest.mark.xfail                   # TODO: remove xfail
    def test_format_index_html(self):
        """Ensure format_index_html works as expected"""
        debug.trace(4, "test_format_index_html()")
        assert(False)

    @pytest.mark.skipif(not TEST_TBD, reason="Ignoring feature to be designed")
    @pytest.mark.xfail                   # TODO: remove xfail
    def test_start_web_controller(self):
        """Ensure start_web_controller works as expected"""
        # Note: pytest has a quirk leading to a logging exception during cherrypy cleanup
        debug.trace(4, "test_start_web_controller()")
        TODO_MODEL = "todo.model"
        wc = THE_MODULE.start_web_controller(TODO_MODEL, nonblocking=True)
        html_listing = wc.index()
        assert("</html>" in html_listing)
        ## TODO2: wc.stop()
        debug.trace(4, "out test_start_web_controller")

class TestTextCategorizerScript(TestWrapper):
    """Class for main testcase definition"""
    script_file = TestWrapper.get_module_file_path(__file__)
    script_module = TestWrapper.get_testing_module_name(__file__, THE_MODULE)
    resources = gh.resolve_path("resources")
    # TODO: use_temp_base_dir = True            # treat TEMP_BASE as directory
    # note: temp_file defined by parent (along with script_module, temp_base, and test_num)

    def test_usage(self):
        """Test usage statement"""
        # Current usage:
        #   Usage: ./text_categorizer.py model
        usage = self.run_script(options="--help")
        assert "model" in usage
    
    @trap_exception
    @pytest.mark.xfail                   # TODO: remove xfail
    def test_test_untrained(self):
        """Make sure accuracy 0 if untrained (TODO: trap stderr)"""
        debug.trace(4, "test_test_untrained()")
        tc = THE_MODULE.TextCategorizer()
        ## OLD: accuracy = tc.test(__file__)
        test_data = ["cat1\ta b c d e",
                     "cat2\tf g h i j"]
        gh.write_lines(self.temp_file, test_data)
        accuracy = tc.test(self.temp_file)
        assert(accuracy == 0)

    @trap_exception
    @pytest.mark.xfail
    def test_train(self):
        """Make sure training works"""
        debug.trace(4, "test_train()")
        data_file = gh.form_path(self.resources, "random-10pct-tweet-emotions.tsv")
        data = system.read_lines(data_file)
        tc = THE_MODULE.TextCategorizer(use_xgb=True)
        tc.train(data_file)
        num_ok = 0
        num_to_test = 10
        for _i in range(num_to_test):
            ## OLD: case = random.randint(1, len(data) - 1)
            case = misc_utils.random_int(1, len(data) - 1)
            expect_cat, expect_text = data[case].split("\t")
            actual_cat = tc.categorize(expect_text)
            debug.trace_expr(5, case, expect_cat, actual_cat)
            if (expect_cat == actual_cat):
                num_ok += 1
        debug.trace_expr(4, num_ok)
        assert(num_ok >= num_to_test // 2)
        

#------------------------------------------------------------------------

if __name__ == '__main__':
    debug.trace_current_context()
    pytest.main([__file__])



#! /usr/bin/env python
#
# Class for text categorizer using Scikit-Learn. See tutorial at
#    http://scikit-learn.org/stable/tutorial/text_analytics/working_with_text_data.html
#
# This includes support for running a few different classifiers. However, for
# more classifiers, panda_sklearn.py should be used, espcially as it includes
# Tensorflow and Keras.
#
# NOTES:
# - Calculates accuracy as number of agreements over number of cases:
#   ex:    (tp + tn) / (tp + tn + fp + fn)    for binary classifications
# - See https://en.wikipedia.org/wiki/Evaluation_of_binary_classifiers#Single_metrics.
# - Also see https://en.wikipedia.org/wiki/Accuracy_and_precision.
# - Keep changes in sync with text_categorizer.py (e.g., XGBoost and GPU options).
# - CherryPy Web server based on following tutorial
#     https://simpletutorials.com/c/2165/How%20to%20Create%20a%20Simple%20JSON%20Service%20with%20CherryPy
#
# TODO:
# - Maintain cache of categorization results.
# - Review categorization code and add examples for clarification of parameters.
# - Fix SHOW_REPORT option for training.
# - Put web server in separate module.
#

"""Text categorization support"""

# Standard packages
import json
from itertools import zip_longest
import os
import re
import sys

# Installed packages
import cherrypy
import numpy
import pandas
from sklearn.base import BaseEstimator, ClassifierMixin
from sklearn.feature_extraction.text import TfidfVectorizer
from sklearn.feature_extraction.text import _document_frequency
from sklearn.naive_bayes import MultinomialNB
from sklearn.linear_model import SGDClassifier
from sklearn.linear_model import LogisticRegression
from sklearn.svm import SVC
from sklearn.pipeline import Pipeline
from sklearn import metrics
from sklearn.utils.multiclass import unique_labels

# Local packages
from mezcla import debug
from mezcla import glue_helpers as gh
from mezcla import misc_utils as misc
from mezcla import system
from mezcla.system import (
    getenv_bool, getenv_float, getenv_int, getenv_text,
)

#................................................................................
# Constants (e.g., environment-based options)

SERVER_PORT = system.getenv_integer("SERVER_PORT", 9010,
                                    "TCP port for web interface")
OUTPUT_BAD = system.getenv_bool("OUTPUT_BAD", False)
CONTEXT_LEN = system.getenv_int("CONTEXT_LEN", 512)
VERBOSE = system.getenv_bool("VERBOSE", False)
OUTPUT_CSV = system.getenv_bool("OUTPUT_CSV", False)
BASENAME = system.getenv_text("BASENAME", "textcat")
OMIT_STOPWORDS = system.getenv_bool("OMIT_STOPWORDS", False,
                                    description="Omit stopwords (e.g., from ngrams)")
CLASS_VAR = getenv_text("CLASS_VAR", "_class_",
                        "Label to use for classification variable")
RETAIN_LABELS = getenv_bool("RETAIN_LABELS", False,
                            "Don't encode class labels")
ENCODE_CLASSES = getenv_bool("ENCODE_CLASSES", not RETAIN_LABELS,
                             "Encode classes using enumeration")
TRACE_IMPORTANCES = getenv_bool("TRACE_IMPORTANCES", False,
                                "Trace feature importances")

# Options for Support Vector Machines (SVM)
#
# Descriptions of the parameters can be found at following page:
#    http://scikit-learn.org/stable/modules/generated/sklearn.svm.SVC.html#sklearn.svm.SVC
# note: defaults used for parameters (n.b., the value None is not usable due to
# sklearn constructor limitations).
USE_SVM = system.getenv_bool("USE_SVM", False)
SVM_KERNEL = system.getenv_text("SVM_KERNEL", "rbf")
SVM_PENALTY = system.getenv_float("SVM_PENALTY", 1.0)
SVM_MAX_ITER = system.getenv_int("SVM_MAX_ITER", -1)
SVM_VERBOSE = system.getenv_bool("SVM_VERBOSE", False)

# Options for Stochastic Gradient Descent (SGD)
#
# Descriptions of the parameters can be found at following page:
#    http://scikit-learn.org/stable/modules/generated/sklearn.linear_model.SGDClassifier.html
# TODO: initialize to None and override only if non-Null
USE_SGD = system.getenv_bool("USE_SGD", False)
SGD_LOSS = system.getenv_text("SGD_HINGE", "hinge")
SGD_PENALTY = system.getenv_text("SGD_PENALTY", "l2")
SGD_ALPHA = system.getenv_float("SGD_ALPHA", 0.0001)
SGD_SEED = system.getenv_float("SGD_SEED", None)
SGD_MAX_ITER = system.getenv_int("SGD_MAX_ITER", 5)
## OLD: SGD_TOLERANCE = system.getenv_float("SGD_TOLERANCE", None)
SGD_VERBOSE = system.getenv_bool("SGD_VERBOSE", False)

# Options for Extreme Gradient Boost (XGBoost)
USE_XGB = system.getenv_bool("USE_XGB", False)
xgb = None
if USE_XGB:
    # pylint: disable=import-outside-toplevel, import-error
    import xgboost as xgb
XGB_BOOSTER = system.getenv_value("XGB_BOOSTER", None)
XGB_USE_GPUS = system.getenv_bool("XGB_USE_GPUS", False)
XGB_VERBOSITY = getenv_int("XGB_VERBOSITY", 0, "Degree of verbosity from 0 to 3")
XGB_JSON = system.getenv_bool("XGB_USE_GPUS", False,
                              "Use XGBoost model in JSON format")
debug.assertion(not XGB_JSON, "JSON support is broke due to obscure manuals")

# Options for Logistic Regression (LR)
# TODO: add regularization
USE_LR = system.getenv_bool("USE_LR", False)

# Options for GPU usage
GPU_DEVICE = system.getenv_value("GPU_DEVICE", None,     # TODO: clarify value to use
                                "Device number for GPU (e.g., shown under nvidia-smi)")

# Options for TFIDF transformation
# Note: these should be None by default to allow for overriding in constructor
# (see TextCategorizer.__init__).
# TODO: add others from sklearn/feature_extraction/text.py
# ex: min/max_df
TFIDF_MAX_TERMS = getenv_int("MAX_TERMS", None,
                             "Maximum number of terms in TF/IDF matrix")
TFIDF_MIN_NGRAM = getenv_int("MIN_NGRAM_SIZE", None)
TFIDF_MAX_NGRAM = getenv_int("MAX_NGRAM_SIZE", None)
TFIDF_MIN_DF = getenv_float("MIN_DF", None)
TFIDF_MAX_DF = getenv_float("MAX_DF", None)
TFIDF_CHAR_NGRAMS = getenv_bool("CHAR_NGRAMS", None)
TFIDF_STOPWORDS = not OMIT_STOPWORDS

# TODO: Options for Naive Bayes (NB), the default
all_use_settings = [USE_SVM, USE_SGD, USE_XGB, USE_LR]
USE_NB = (not any(all_use_settings))
debug.assertion(sum(int(use) for use in all_use_settings) <= 1)

#................................................................................
# Utility functions

def sklearn_report(actual, predicted, actual_labels, predicted_labels, stream=sys.stdout):
    """Print classification analysis report for ACTUAL vs. PREDICTED indices with original LABELS and using STREAM"""
    stream.write("Performance metrics:\n")
    indices = unique_labels(actual, predicted)
    labels = unique_labels(actual_labels, predicted_labels)
    stream.write(metrics.classification_report(actual, predicted,
                                               labels=indices, target_names=labels))
    stream.write("Confusion matrix:\n")
    # TODO: make showing all cases optional
    confusion = metrics.confusion_matrix(actual, predicted,
                                         labels=indices)
    # TODO: make sure not clipped
    stream.write("{cm}\n".format(cm=confusion))
    debug.trace_object(6, confusion, "confusion")
    return


def create_tabular_file(filename, data):
    """Create tabular FILENAME with SkLearn DATA for use with read_categorization_data"""
    # Note: intended for comparing results here against tutorial (e.g., in ipython shell)
    with system.open_file(filename, "w") as f:
        for i in range(len(data.data)):
            text = system.to_utf8(re.sub("[\t\n]", " ", data.data[i]))
            f.write("{lbl}\t{txt}\n".format(lbl=data.target_names[data.target[i]], txt=text))
    return


def read_categorization_data(filename):
    """Reads table with (non-unique) label and tab-separated value. 
    Note: label made lowercase; result returned as tuple (labels, values)"""
    debug.trace_fmtd(4, "read_categorization_data({f})", f=filename)
    labels = []
    values = []
    with system.open_file(filename) as f:
        for (i, line) in enumerate(f):
            line = system.from_utf8(line)
            items = line.split("\t")
            if len(items) == 2:
                labels.append(items[0].lower())
                values.append(items[1])
            else:
                debug.trace_fmtd(4, "Warning: Ignoring item w/ unexpected format at line {num}: items: len={l} first={f} second={s}",  l=len(items), f=gh.elide(items[0]), s=gh.elide(items[0]),
                                 num=(i + 1))
    debug.trace_values(7, zip(labels, values), "table")
    return (labels, values)


def int_if_whole(num):
    """Return integral type if NUM is a whole number"""
    # EX: int_if_whole(2.0) => 2
    # EX: int_if_whole(2.999) => 2.999
    result = int(num) if (round(num) == num) else num
    debug.trace_fmt(7, "int_if_whole({n}) => {r}", n=num, r=result)
    return result


def param_or_default(param_value, default_value):
    """Returns PARAM_VALUE if non-null else DEFAULT_VALUE"""
    # Note: this avoid awkward initialization sequences in constructor
    # - verbose version:
    #   if param_value in None:
    #      param_value = default_value
    #   self.member = param_value
    # - long-line version
    #   self.member = param_value if (param_value is not None) else default_value
    # - streamlined version
    #   self.member = param_or_default(param, default)
    result = param_value if (param_value is not None) else default_value
    return result

#...............................................................................

class ClassifierWrapper(BaseEstimator, ClassifierMixin):
    """Wrapper around arbitrary categorizer object, originally used for the sake of tracing the feature vectors derived via pipelines"""
    ## TODO: rework by specializing one classifier (e.g., MultinomialNB) so that fit() only needs to be defined

    def __init__(self, classifier):
        """Constructor: records CLASSIFIER"""
        debug.trace_fmt(6, "{cl}.__init__(clf={c})", c=classifier, cl=str(type(self)))
        self.classifier = classifier
        self.tfidf_vectorizer = None

    def _get_param_names(self):
        """Get parameter names for the estimator"""
        # TODO: drop method
        # Note: This is not class method as in BaseEstimator.
        # pylint: disable=protected-access,arguments-differ
        return self.classifier._get_param_names()

    def get_params(self, deep=True):
        """Return list of parameters supported"""
        return self.classifier.get_params(deep=deep)

    def fit(self, training_x=None, training_y=None):
        """Delegates fit() invocation to classifier, after outputing CSV if desired"""
        if OUTPUT_CSV:
            self.output_csv(training_x, training_y, BASENAME)
        fit_result = self.classifier.fit(training_x, training_y)
        if TRACE_IMPORTANCES:
            debug.trace_fmt(1, "Feature importances: {imp}",
                            imp=self.extract_feature_importance())
        ## DEBUG: debug.trace_object(6, self.tfidf_vectorizer)
        return fit_result

    def predict(self, sample):
        """Return predicted class for each SAMPLE (returning vector)"""
        if OUTPUT_CSV:
            self.output_csv(sample, (["n/a"] * sample.shape[0]), BASENAME + ".test")
        return self.classifier.predict(sample)

    def score(self, X, y, sample_weight=None):
        """Return average accuracy of prediction of X matrix vs Y vector"""
        return self.classifier.score(X, y, sample_weight=sample_weight)
    
    def predict_proba(self, sample):
        """Return probabilities of outcome classes predicted for each SAMPLE (returning matrix)"""
        return self.classifier.predict_proba(sample)

    def output_csv(self, x, y, basename):
        """Output features in X and Y to BASENAME.csv"""
        # TODO: move into TextCategorizer; put pickle and IDF support in separate method
        debug.trace(6, f"output_csv(_, _, {basename}); self={self}")
        debug.reference_var(self)

        ## NOTE: save in pickle format for debugging
        if debug.verbose_debugging():
            system.save_object(basename + ".x.csv.pickle", x)
            system.save_object(basename + ".y.csv.pickle", y)
        ##
        df_x = pandas.DataFrame(x.toarray())
        df_y = pandas.DataFrame(y)
        
        ## HACK: use global pipeline to get feature names
        def normalize(feature_name):
            """Normalize FEATURE_NAME"""
            return feature_name.replace(" ", "_")
        ##
        features = [normalize(f) for f in self.tfidf_vectorizer.get_feature_names()]
        df_x.to_csv(basename + ".x.csv.list", header=features, index=False)
        df_y.to_csv(basename + ".y.csv.list", header=[CLASS_VAR], index=False)
        gh.run("paste --delimiters=',' {b}.x.csv.list {b}.y.csv.list > {b}.csv.list",
               b=basename)
        # TODO: combine into single dataframe and use to_csv over that; use .csv instead of .csv.list (and make sure not to overwrite)
        debug.assertion(system.file_exists(basename + ".csv.list"))

        # Trace out the IDF values
        ## TODO: debug.trace_value(6, zip(features, tfidf_vectorizer._idf_diag), "ngram idf's")
        if debug.debugging(6):
            # pylint: disable=protected-access
            df = _document_frequency(x)
            sorted_features = sorted(zip(features, self.tfidf_vectorizer.idf_, df),
                                     key=lambda f_idf: f_idf[1], reverse=True)
            debug.trace_values(6, sorted_features, "ngram idf's & df's")
        return
    
    def extract_feature_importance(self):
        """Returns list of (name, weight) for important features"""
        # TODO: move into TextCategorizer
        debug.trace(6, f"extract_feature_importance(); self={self}")
        result = []
        try:
            feature_names = (self.tfidf_vectorizer.get_feature_names() or [])
            sorted_scores = sorted(zip_longest(feature_names, self.classifier.feature_importances_,  fillvalue="F?"),
                                   key=lambda name_score: name_score[1],
                                   reverse=True)
            result = [(f, s) for (f, s) in sorted_scores if s > 0]
        except:
            system.print_exception_info("extract_feature_importance")
        debug.trace(5, f"extract_feature_importance() => {result}")
        return result

#...............................................................................

class TextCategorizer(object):
    """Class for building text categorization"""
    # TODO: add cross-fold validation support; make TF/IDF weighting optional

    def __init__(self, tfidf_max_terms=None, tfidf_min_ngram=None, tfidf_max_ngram=None, tfidf_min_df=None, tfidf_max_df=None, tfidf_stopwords=None, tfidf_char_ngrams=None,
                 use_xgb=None, use_svm=None, use_sgd=None, use_lr=None):
        """Class constructor: initializes classifier and text categoriation pipeline"""
        debug.trace_fmtd(4, "tc.__init__(); self={s}", s=self)
        self.tfidf_max_terms = param_or_default(tfidf_max_terms, TFIDF_MAX_TERMS)
        self.tfidf_min_ngram = param_or_default(tfidf_min_ngram, TFIDF_MIN_NGRAM)
        self.tfidf_max_ngram = param_or_default(tfidf_max_ngram, TFIDF_MAX_NGRAM)
        self.tfidf_min_df = param_or_default(tfidf_min_df, TFIDF_MIN_DF)
        self.tfidf_max_df = param_or_default(tfidf_max_df, TFIDF_MAX_DF)
        self.tfidf_stopwords = param_or_default(tfidf_stopwords, TFIDF_STOPWORDS)
        self.tfidf_char_ngrams = param_or_default(tfidf_char_ngrams, TFIDF_CHAR_NGRAMS)
        ## TODO: self.member = param_or_default(xyz, XYZ)
        #
        self.keys = []
        self.classifier = None
        classifier = None
        if use_xgb is None:
            use_xgb = USE_XGB
        if use_svm is None:
            use_svm = USE_SVM
        if use_sgd is None:
            use_sgd = USE_SGD
        if use_lr is None:
            use_lr = USE_LR

        # Derive classifier based on user options
        if use_svm:
            classifier = SVC(kernel=SVM_KERNEL,
                             C=SVM_PENALTY,
                             max_iter=SVM_MAX_ITER,
                             verbose=SVM_VERBOSE)
        elif use_sgd:
            classifier = SGDClassifier(loss=SGD_LOSS,
                                       penalty=SGD_PENALTY,
                                       alpha=SGD_ALPHA,
                                       random_state=SGD_SEED,
                                       ## TODO: max_iter=SGD_MAX_ITER,
                                       verbose=SGD_VERBOSE)
            ## HACK: support old version and new (thanks sklearn!)
            ## TODO: make sure this won't break (e.g., due to visibility)
            num_iter_attribute = "max_iter"
            if (not hasattr(classifier, num_iter_attribute)):
                num_iter_attribute = "n_iter"
            debug.assertion(hasattr(classifier, num_iter_attribute))
            setattr(classifier, num_iter_attribute, SGD_MAX_ITER)
        elif use_xgb:
            # TODO: rework to just define classifier here and then pipeline at end.
            # in order to eliminate redundant pipeline-specification code.
            # TODO: n_jobs=-1
            misc_xgb_params = {'booster': XGB_BOOSTER, 'verbosity': XGB_VERBOSITY}
            if XGB_USE_GPUS:
                misc_xgb_params.update({'tree_method': 'gpu_hist'})
                misc_xgb_params.update({'predictor': 'gpu_predictor'})
            if GPU_DEVICE:
                misc_xgb_params.update({'gpu_id': GPU_DEVICE})
            debug.trace_fmt(4, 'misc_xgb_params={m}', m=misc_xgb_params)
            classifier = xgb.XGBClassifier(**misc_xgb_params)
        elif use_lr:
            classifier = LogisticRegression()
        else:
            debug.assertion(USE_NB)
            classifier = MultinomialNB()
        use_classifier_wrapper = (OUTPUT_CSV or TRACE_IMPORTANCES)
        if use_classifier_wrapper:
            classifier = ClassifierWrapper(classifier)
            debug.trace_fmt(4, "Using wrapper ({cl}) for feature tracing hooks", cl=type(classifier))

        # Add classifier to text categorization pipeline]
        tfidf_parameters = {}
        if self.tfidf_max_terms:
            ## TODO: sort by TF/IDF (not TF)
            tfidf_parameters['max_features'] = self.tfidf_max_terms
        if self.tfidf_min_ngram or self.tfidf_max_ngram:
            min_ngram = (self.tfidf_min_ngram or 1)
            max_ngram = (self.tfidf_max_ngram or 1)
            debug.assertion(1 <= min_ngram <= max_ngram)
            tfidf_parameters['ngram_range'] = (min_ngram, max_ngram)
        if self.tfidf_min_df:
            tfidf_parameters['min_df'] = int_if_whole(self.tfidf_min_df)
        if self.tfidf_max_df:
            tfidf_parameters['max_df'] = int_if_whole(self.tfidf_max_df)
        if not self.tfidf_stopwords:
            tfidf_parameters['stop_words'] = 'english'
        if self.tfidf_char_ngrams:
            tfidf_parameters['analyzer'] = 'char'
        self.cat_pipeline = Pipeline(
            [('tfidf', TfidfVectorizer(**tfidf_parameters)),
             ('clf', classifier)])
        if use_classifier_wrapper:
            pipeline_steps = list(self.cat_pipeline._iter())
            debug.assertion(pipeline_steps[0][1] == 'tfidf')
            ## TODO: classifier.tfidf_vectorizer = self.cat_pipeline['tfidf']
            classifier.tfidf_vectorizer = pipeline_steps[0][2]
        debug.trace_object(5, self, "TextCategorizer")
        return

    def train(self, filename):
        """Train classifier using tabular FILENAME with label and text"""
        debug.trace_fmtd(4, "tc.train({f})", f=filename)
        (labels, values) = read_categorization_data(filename)
        label_values = labels
        self.keys = sorted(numpy.unique(labels))
        debug.trace_expr(5, self.keys)
        if ENCODE_CLASSES:
            label_indices = [self.keys.index(l) for l in labels]
            label_values = label_indices
        self.classifier = self.cat_pipeline.fit(values, label_values)
        debug.trace_object(7, self, "TextCategorizer")
        return

    def test(self, filename, report=False, stream=sys.stdout):
        """Test classifier over tabular data from FILENAME with label and text, returning accuracy. Optionally, a detailed performance REPORT is output to STREAM."""
        debug.trace_fmtd(4, "tc.test({f})", f=filename)
        (all_labels, all_values) = read_categorization_data(filename)
        debug.trace_values(6, all_labels, "all_labels")
        debug.trace_values(6, [gh.elide(v) for v in all_values], "all_values")

        # Prune cases with classes not in training data
        # TODO: use hash of positions
        actual_indices = []
        values = []
        labels = []
        for (i, label) in enumerate(all_labels):
            if label in self.keys:
                values.append(all_values[i])
                actual_indices.append(self.keys.index(label))
                labels.append(label)
            else:
                debug.trace_fmtd(4, "Ignoring test label {l} not in training data (line {n})",
                                 l=label, n=(i + 1))

        # Perform classification and determine accuracy
        try:
            predicted_values = self.classifier.predict(values)
            if ENCODE_CLASSES:
                predicted_indices = predicted_values
            else:
                predicted_indices = [self.keys.index(label) for label in predicted_values]
            debug.assertion(len(actual_indices) == len(predicted_indices))
            debug.trace_values(6, actual_indices, "actual")
            debug.trace_values(6, predicted_indices, "predicted")
            ## TODO: predicted_labels = [self.keys[i] for i in predicted_indices]
            num_ok = sum((actual_indices[i] == predicted_indices[i]) for i in range(len(actual_indices)))
            accuracy = float(num_ok) / len(values)
        except:
            accuracy = 0
            system.print_exception_info("tc.test")

        # Output classification report
        if report:
            debug.assertion(VERBOSE)
            if VERBOSE:
                stream.write("Missed classifications")
                stream.write("\n")
                stream.write("Actual\tPredict\n")
                ## TODO: complete conversion to using actual_index (here and below)
                num_missed = 0
                for (i, actual_index) in enumerate(actual_indices):
                    debug.assertion(actual_index == actual_indices[i])     # pylint: disable=unnecessary-list-index-lookup
                    if (actual_index != predicted_indices[i]):
                        stream.write("{act}\t{pred}\n".
                                     format(act=self.keys[actual_index],
                                            pred=self.keys[predicted_indices[i]]))
                        num_missed += 1
                if (num_missed == 0):
                    stream.write("n/a")
                stream.write("\n")
            actual_labels = [self.keys[i] for i in actual_indices]
            predicted_labels = [self.keys[i] for i in predicted_indices]
            sklearn_report(actual_indices, predicted_indices, actual_labels, predicted_labels, stream)

        # Show cases not classified OK
        if OUTPUT_BAD:
            bad_instances = "Actual\tBad\tText\n"
            # TODO: for (i, actual_index) in enumerate(actual_indices)
            for (i, actual_index) in enumerate(actual_indices):
                debug.assertion(actual_index == actual_indices[i])   # pylint: disable=unnecessary-list-index-lookup
                if (actual_index != predicted_indices[i]):
                    text = values[i]
                    context = (text[:CONTEXT_LEN] + "...\n") if (len(text) > CONTEXT_LEN) else text
                    # TODO: why is pylint flagging the format string as invalid?
                    bad_instances += "{g}\t{b}\t{t}\n".format(
                        g=self.keys[actual_index],
                        b=self.keys[predicted_indices[i]],
                        t=context)
            bad_filename = filename + ".bad"
            system.write_file(bad_filename, bad_instances)
            debug.trace_fmt(4, "Result ({f}):\n{r}", f=bad_filename, r=system.read_file(bad_filename))
        return accuracy

    def categorize(self, text):
        """Return category for TEXT"""
        debug.trace(4, "tc.categorize(_)")
        debug.trace_fmtd(6, "\ttext={t}", t=text)
        try:
            index = self.classifier.predict([text])[0]
            label = self.keys[index]
        except:
            system.print_exception_info("categorize")
        debug.trace_fmtd(6, "categorize() => {r}", r=label)
        return label

    def class_probabilities(self, text):
        """Return probability distribution for TEXT"""
        debug.trace(4, "tc.class_probabilities(_)")
        debug.trace_fmtd(6, "\ttext={t}", t=text)
        dist = None
        try:
            class_names = self.keys
            class_probs = self.classifier.predict_proba([text])[0]
            debug.trace_object(7, self.classifier)
            debug.trace_fmtd(6, "class_names: {cn}\nclass_probs: {cp}", cn=class_names, cp=class_probs)
            sorted_scores = misc.sort_weighted_hash(dict(zip(class_names, class_probs)))
            dist = " ".join([(k + ": " + system.round_as_str(s)) for (k, s) in sorted_scores])
        except:
             system.print_exception_info("class_probabilities")
        debug.trace_fmtd(5, "class_probabilities() => {r}", r=dist)
        return dist

    def save(self, filename):
        """Save classifier to FILENAME
        Note: with XGB_JSON, the XGBoost JSON format is used (for better portability).
        """
        debug.trace_fmtd(4, "tc.save({f})", f=filename)
        try:
            # pylint: disable=no-value-for-parameter, no-else-raise, unreachable
            if XGB_JSON:
                raise NotImplementedError()
                xgb.XGBModel.save_model(filename)
                ## TODO: get xgboost to save the keys in the model JSON file
                system.write_file(filename + ".keys", json.dumps(self.keys))
            else:
                system.save_object(filename, [self.keys, self.classifier])
        except:
            system.print_exception_info("tc.save")
        return

    def load(self, filename):
        """Load classifier from FILENAME
        Note: with XGB_JSON, uses the XGBoost JSON format.
        """
        debug.trace_fmtd(4, "tc.load({f})", f=filename)
        try:
            # pylint: disable=no-value-for-parameter, no-else-raise, unreachable, assignment-from-none
            if XGB_JSON:
                raise NotImplementedError()
                ## TODO2: fix assignment
                self.classifier = xgb.XGBModel.load_model(filename)
                ## HACK: load keys separately
                self.keys = json.loads(system.read_file(filename + ".keys"))
            else:
                (self.keys, self.classifier) = system.load_object(filename)
        except (TypeError, ValueError):
            system.print_stderr("Problem loading classifier from {f}: {exc}".
                                format(f=filename, exc=sys.exc_info()))
        return

#-------------------------------------------------------------------------------
# CherryPy Web server based on following tutorial
#     https://simpletutorials.com/c/2165/How%20to%20Create%20a%20Simple%20JSON%20Service%20with%20CherryPy
#

# Constants
TRUMP_TEXT = "Donald Trump was President."
DOG_TEXT = "My dog has fleas."

#--------------------------------------------------------------------------------
# Utility function(s)

def format_index_html(base_url=None):
    """Formats a simple HTML page illustrating the categorize and class_probabilities API calls,
    Note: BASE_URL provides the server URL (e.g., http://www.my-categorizer.com:9999)"""
    debug.trace(5, f"format_index_html(base_url={base_url})")
    # TODO: parameterize template generation (e.g., to facilitate usage in derived classes of web_controller
    if (base_url is None):
        base_url = "http://127.0.0.1"
    if (base_url.endswith("/")):
        base_url = system.chomp(base_url, "/")

    # Create index page template with optional examples for debugging
    html_template = """
    <!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.01 Transitional//EN">
    <html lang="en">
        <head>
            <meta content="text/html; charset=UTF-8" http-equiv="content-type">
            <title>Text categorizer</title>
        </head>
        <body>
            Try <a href="categorize">categorize</a> and <a href="class_probabilities">class_probabilities</a>.<br>
            note: You need to supply the <i><b>text</b></i> parameter.<br>
            <br>
            Examples:
            <ul>
                <li>Category for <a href="categorize?text={quoted_trump_text}">"{trump_text}"</a>:<br>
                    {indent}<code>{base_url}/categorize?text={quoted_trump_text}</code>
                </li>
    
                <li>Probability distribution for <a href="class_probabilities?text={quoted_dog_text}">"{dog_text}"</a>:<br>
                    {indent}<code>{base_url}/class_probabilities?text={quoted_dog_text}</code>
                </li>
            </ul>
    """

    if debug.detailed_debugging():
        html_template += """
            <!-- <p> -->
            Other examples (n.b., debug only):
            <ul>
                <li><a href="shutdown">Shutdown</a> the server:<br>
                    {indent}<code>{base_url}/shutdown</code>
                </li>

                <li>Alias for <a href="index">this index page</a>:<br>
                    {indent}<code>{base_url}/index</code>
                </li> 
            </ul>
        """
    #
    # TODO: define text area dimensions based on browser window size
    html_template += """
            <!-- Form for entering text for categorization -->
            <hr>
            <form action="{base_url}/categorize" method="get">
                <label for="textarea1">Categorize</label>
                <br>
                <textarea id="textarea1" rows="10" cols="100" name="text"></textarea>
                <br>
                <input type="submit">
            </form>            
    """
    #
    html_template += """
            <!-- Form for entering text for textcat probability distribution -->
            <hr>
            <form action="{base_url}/probs" method="get">
                <label for="textarea1">Probabilities</label>
                <br>
                <textarea id="textarea2" rows="10" cols="100" name="text"></textarea>
                <br>
                <input type="submit">
            </form>
    """
    html_template += """          
        </body>
    </html>
    """

    # Resolve template into final HTML
    index_html = html_template.format(base_url=base_url,
                                      indent="&nbsp;&nbsp;&nbsp;&nbsp;",
                                      trump_text=TRUMP_TEXT,
                                      quoted_trump_text=system.quote_url_text(TRUMP_TEXT),
                                      dog_text=DOG_TEXT,
                                      quoted_dog_text=system.quote_url_text(DOG_TEXT))
    return index_html

#................................................................................
# Main class

class web_controller(object):
    """Controller for CherryPy web server with embedded text categorizer"""
    
    def __init__(self, model_filename, *args, **kwargs):
        """Class constructor: initializes search engine server"""
        debug.trace_fmtd(5, "web_controller.__init__(s:{s}, a:{a}, kw:{k})__",
                         s=self, a=args, k=kwargs)
        self.text_cat = TextCategorizer()
        self.text_cat.load(model_filename)
        return

    @cherrypy.expose
    def index(self, **kwargs):
        """Website root page (e.g., web site overview and link to search)"""
        # TODO: add way to override URL (e.g., to force use of known local hostname instead of "localhost")
        debug.trace_fmtd(5, "wc.index(s:{s}, kw:{kw})", s=self, kw=kwargs)
        base_url = cherrypy.url('/')
        debug.trace_fmt(4, "base_url={b}", b=base_url)
        index_html = format_index_html(base_url)
        debug.trace_fmt(6, "html={{\n{h}\n}}", h=index_html)
        return index_html

    @cherrypy.expose
    def categorize(self, text, **kwargs):
        """Infer category for TEXT"""
        debug.trace_fmtd(5, "wc.categorize(s:{s}, _, kw:{kw})", s=self, kw=kwargs)
        return self.text_cat.categorize(text)

    @cherrypy.expose
    def class_probabilities(self, text, **kwargs):
        """Get category probability distribution for TEXT"""
        debug.trace_fmtd(5, "wc.class_probabilities(s:{s}, _, kw:{kw})", s=self, kw=kwargs)
        return self.text_cat.class_probabilities(text)
    #
    probs = class_probabilities

    @cherrypy.expose
    def stop(self, **kwargs):
        """Stops the web search server and saves cached data to disk.
        Note: The command is ignored if not debugging and on a production server."""
        debug.trace_fmtd(5, "wc.stop(s:{s}, kw:{kw})", s=self, kw=kwargs)
        # TODO: replace stooges with your real server nicknames
        if ((not debug.detailed_debugging()) and (os.environ.get("HOST_NICKNAME") in ["curly", "larry", "moe"])):
            return "Call security!"
        # TODO: Straighten out shutdown quirk (seems like two invocations required).
        # NOTE: Putting exit before stop seems to do the trick. However, it might be
        # the case that the server shutdown.
        cherrypy.engine.exit()
        cherrypy.engine.stop()
        # TODO: Use HTML so shutdown shown in title.
        return "Adios"

    # alias for stop
    shutdown = stop
    # TODO: track down delay in python process termination


def start_web_controller(model_filename, nonblocking=False):
    """Start up the CherryPy controller for categorization via MODEL_FILENAME
    Note:
    - The function blocks until server is shutdown unless NONBLOCKING specified.    """
    debug.trace(5, "start_web_controller()")

    # Load in CherryPy configuration
    # TODO: use external configuration file
    conf = {
        '/': {
            'tools.sessions.on': True,
            'tools.staticdir.root': os.path.abspath(os.getcwd()),
            ## notes: avoids cross-origin type errrors
            'tools.response_headers.on': True,
            'tools.response_headers.headers': [
                ## TODO: just allow the same host
                ('Access-Control-Allow-Origin', '*'),
            ]
        },
        'global': {
            'server.socket_host': "0.0.0.0",
            'server.socket_port': SERVER_PORT,
            'server.thread_pool': 10,
            }
        }

    # Start the server
    # TODO: trace out all configuration settings
    debug.trace_values(4, cherrypy.response.headers, "default response headers")
    textcat_controller = web_controller(model_filename)
    debug.trace_expr(4, textcat_controller.text_cat.keys)
    ## OLD:
    ## cherrypy.quickstart(textcat_controller, config=conf)
    ## # Note: the following call blocks
    ## cherrypy.engine.start()
    if nonblocking:
        # Based on quickstart in cherrypy/__init__.py (version 18.7.0), sans the block:
        cherrypy._global_conf_alias.update(conf)  # pylint: disable=protected-access
        cherrypy.tree.mount(textcat_controller, config=conf)
        cherrypy.engine.signals.subscribe()
        debug.trace_expr(4, "starting cherrypy server")
        cherrypy.engine.start()
    else:
        # Note: the following call blocks
        debug.trace_expr(4, "quick-starting cherrypy server")
        cherrypy.quickstart(textcat_controller, config=conf)
    return textcat_controller

#------------------------------------------------------------------------
# Entry point

def main(args):
    """Supporting code for command-line processing"""
    debug.trace_fmtd(6, "main({a})", a=args)
    # HACK: ignore --tag label (n.b., used for killing via process regex)
    ## BAD: if ((len(args) > 0) and (args[1] == "--tag")):
    if ((len(args) > 1) and (args[1] == "--tag")):
        args[1:] = args[3:]
    if ((len(args) != 2) or (args[1] == "--help")):
        print("Usage: {p} model".format(p=args[0]))
        return
    model = args[1]
    start_web_controller(model)
    return

if __name__ == '__main__':
    main(sys.argv)



#! /usr/bin/env python
#
# BATSPP
#
# This process and run custom tests using bats-core.
#
# NOTE: It is only necessary to have installed bats-core
#       no need for bats-assertions library.
#
# FYI: This version is a simplified version of the original BatsPP facility. It uses some shameless
# hacks enabled via environment variables to work around quirks in the example parsing. For example,
#   TEST_FILE=1 MATCH_SENTINELS=1 PARA_BLOCKS=1 python ./simple_batspp.py tests/adhoc-tests.test
#
## TODO:
## - extend usage guide or docstring.
## - Add some directives in the test or script comments:
##        Block of commands used just to setup test and thus without output or if the output should be ignored (e.g., '# Setup').
##        Tests that should be evaluated in the same environment as the preceding code (e.g., '# Continuation'). For example, you could have multiple assertions in the same @test function.
## - find regex match excluding indent directly.
## - pretty result.
## - setup for functions tests.
## - multiline commands?.
## - solve comma sanitization, test poc [TODO: what is the issue?]:
##        $ BATCH_MODE=1 bash -i -c 'source ../tomohara-aliases.bash; mkdir -p $TMP/test-$$; cd $TMP/test-$$; touch  F1.txt F2.list F3.txt F4.list F5.txt; ls | old-count-exts'
##        .txt\t3
##        .list\t2
## - add a tag to avoid running a certain test (example "# OLD").
## - add option for runnng generated bats code through shellcheck
##   This requires that the test definitions be converted to proper bash functions:
##      perl -pe 's/^\@test "(.*)"/function $1/;' my-test.bats > my-text.bash
## - add examples below to help clarify processing
##

"""
BATSPP

This process and run custom tests using bats-core.

You can run tests for aliases and more using the
command line with '$ [command]' followed by the
expected output:

 $ echo -e "hello\nworld"
 hello
 world

 $ echo "this is a test" | wc -c
 15

Also you can test bash functions:
 [functions + args] => [expected]:
 fibonacci 9 => "0 1 1 2 3 5 8 13 21 34"
"""


# Standard packages
from collections import namedtuple
import os
import re
import random


# Local packages
from mezcla.main import Main
from mezcla.my_regex import my_re
from mezcla      import system
from mezcla      import debug
from mezcla      import glue_helpers as gh


#-------------------------------------------------------------------------------

# Command-line labels constants
TESTFILE = 'testfile' # target test path
OUTPUT   = 'output'   # output BATS test
VERBOSE  = 'verbose'  # show verbose debug
SOURCE   = 'source'   # source another file

# Environment options
TMP = system.getenv_text("TMP", "/tmp",
                         "Temporary directory")
TEMP_DIR =  system.getenv_text("TEMP_DIR", gh.form_path(TMP, f"batspp-{os.getpid()}"),
                               "Temporary directory to use for tests")
COPY_DIR = system.getenv_bool("COPY_DIR", False,
                              "Copy current directory to temp. dir for input files, etc.")
# Options to work around quirks with Batspp
PREPROCESS_BATSPP = system.getenv_bool("PREPROCESS_BATSPP", False,
                                       "Preprocess .batspp format file, removing line continuations")
MATCH_SENTINELS = system.getenv_bool("MATCH_SENTINELS", False,
                                     "Include test comment header and trailer in match")
PARA_BLOCKS = system.getenv_bool("PARA_BLOCKS", False,
                                 "Test definitions within perl-style paragraphs--\n\n ends")
# Other useful options
TRACE_MATCHING = system.getenv_bool("TRACE_MATCHING", False,
                                    "Trace out test matching for text")
RANDOM_ID = system.getenv_bool("RANDOM_ID", False,
                               "Use random ID's for tests")
OMIT_PATH = system.getenv_bool("OMIT_PATH", False,
                               "Omit PATH spec. for directory of .Batspp file")
BATS_OPTIONS = system.getenv_text("BATS_OPTIONS", " ",
                                  "Options for bats command")
SKIP_BATS = system.getenv_bool("SKIP_BATS", False,
                               "Do not run the bats test script")
OMIT_TRACE = system.getenv_bool("OMIT_TRACE", False,
                                "Omit actual/expected trace from bats file")
OMIT_MISC = system.getenv_bool("OMIT_MISC", False,
                               "Omit miscellaenous/obsolete bats stuff")
TEST_FILE = system.getenv_bool("TEST_FILE", False,
                               "Treat input as example-base test file, not a bash script")
#
# Shameless hacks
## TEST: GLOBAL_SETUP = system.getenv_text("GLOBAL_SETUP", " ",
##                                   "Global setup bash snippet")
EXTRACT_SETUP = system.getenv_bool("EXTRACT_SETUP", False,
                                   "Extract setup based on '# Setup' comment in entire match")
AUGMENT_COMMANDS = system.getenv_bool("AUGMENT_COMMANDS", False,
                                      "Add commands missing from setup or actual from entire match")
DISABLE_ALIASES = system.getenv_bool("DISABLE_ALIASES", False,
                                      "Disable alias expansion")
MERGE_CONTINUATION = system.getenv_bool("MERGE_CONTINUATION", False,
                                        "Merge function or backslash continuations in expected with actual")
# Flags
USE_INDENT_PATTERN = system.getenv_bool("USE_INDENT_PATTERN", False,
                                        "Use old regex for indentation")

# Some constants
## Bruno: can you explain this pattern?
INDENT_PATTERN = r'^[^\w\$\(\n\{\}]'               # note: modified in __process_tests
BATSPP_EXTENSION = '.batspp'
# Trace levels either go from 4 .. 7 or from 6 .. 9
T6 = system.getenv_int("T6", 6,
                       "Trace level to use for T6")
T7 = (T6 + 1)
T8 = (T7 + 1)
T9 = (T8 + 1)

## TEMP
# pylint: disable=f-string-without-interpolation

#-------------------------------------------------------------------------------
# Utility functions

def merge_continuation(actual, expected):
    """Merge EXPECTED into ACTUAL if line continuation (e.g., function definition)
    Note: returns tuple with new actual and expected

    >>> merge_continuation(
    ...      '''
    ...          function f {
    ...      ''', 
    ...      '''
    ...            x=1
    ...          }
    ...
    ...          123
    ...      ''')
    (
      '''
         function f {
           x=1
         }
      ''', 
      '''
         123
      ''')
    """
    ## ex: merge_continuation("function g () {}\n", "123\n") => ("function g () {}\n", "123\n")  # no change
    actual_lines = actual.split("\n")
    expected_lines = expected.split("\n")

    # Check for open function definition or trailing backslash at end of actual section
    line_continuation = False
    function_continuation = False
    for s in range(len(actual_lines) - 1, -1, -1):
        actual_line = actual_lines[s]
        if re.search(r"function.*{\s*$", actual_line):
            debug.trace(T6, f"Open function definition at actual line {s + 1}: {actual_line!r}")
            function_continuation = True
            break
        if re.search(r"[^\\]\\(\n?)$", actual_line):
            debug.trace(T6, "Line continuation at actual line {s + 1}: {actual_line!r}")
            line_continuation = True
            break
        debug.trace(T9, f"Non-continuation at actual line {s + 1}: {actual_line!r}")

    # If during continuuation, merge lines until no longer at a continuation
    for c, expected_line in enumerate(expected_lines):
        if line_continuation or function_continuation:
            debug.trace(T8, f"Merging expected line {c + 1} with actual: {expected_line!r}")
            actual_lines.append(expected_line)
            expected_lines = expected_lines[1:]
        else:
            break
        line_continuation = re.search(r"[^\\]\\(\n?)", expected_line)
        function_continuation = (function_continuation and (not re.search(r"^\s+\}", expected_line)))
        debug.trace_expr(T7, expected_line, line_continuation, function_continuation)

    # Sanity check
    new_actual = "\n".join(actual_lines)
    new_expected = "\n".join(expected_lines)
    debug.assertion((new_actual + new_expected) == (actual + expected))
    new = ("new" if (new_actual != actual) else "old")
    debug.trace(T9, f"merge_continuation({actual!r}, {expected!r})) => {new} ({new_actual!r}, {new_expected!r})")
    return (new_actual, new_expected)


def preprocess_batspp(contents):
    """Preprocess the file CONTENTS in the .batspp format (e.g., remove line continuations)
    
    >>> preprocess_batspp(r'''
    ... function fu {\
    ...   echo "fu"; \
    ... }
    ... ''')
    '''function fu { echo "fu"; }'''
    """
    # ex: preprocess_batspp('function bar { echo "bar"; }') => 'function bar { echo "bar"; }'   # no change
    # Note: this removes line continuations for sake of simpler parsing (e.g., avoid merge_continuation)
    debug.trace(T6, "preprocess_batspp(_)")
    ## Bruno: can you fix this regex replacement?
    ## TODO: new_contents = my_re.sub(r"^(.*[^\\])\\(\n)$", r"\1\n", contents, flags=re.MULTILINE)
    lines = contents.split("\n")
    ## OLD: for s in range(len(lines) - 1):
    s = 0
    while (s < len(lines)):
        line = lines[s]
        debug.trace(T7, f"checking line {s + 1}: {line}")
        if ((line is not None) and line.endswith("\\") and (len(line) > 1) and (line[-2] != "\\") and (s < len(lines))):
            debug.trace(T6, f"joining lines {s + 1} and {s + 2}")
            lines[s] = line[:-1] + lines[s + 1]
            for j in range(s + 1, len(lines) - 1):
                lines[j] = lines[j + 1]
            lines[-1] = None
        else:
            s += 1
    new_contents = "\n".join(l for l in lines if (l is not None))
    new = ("new" if (new_contents != contents) else "old")
    debug.trace(T8, f"preprocess_batspp({contents!r}) => {new} {new_contents!r}")
    debug.trace(T8, f"\tlen(c)={len(contents)}; len(nc)={len(new_contents)}")
    debug.assertion(len(new_contents) <= len(contents))
    return new_contents

#-------------------------------------------------------------------------------

TEST_FIELD_NAMES = ["entire", "title", "setup", "actual", "expected"]

TestFieldTypes = namedtuple("TestFieldTypes", TEST_FIELD_NAMES)

                                                
class Batspp(Main):
    """This process and run custom tests using bats-core"""


    # Class-level member variables for arguments (avoids need for class constructor)
    testfile     = ''
    output       = ''
    source       = ''
    verbose      = False


    # Global States
    is_test_file = False
    file_content = ''
    bats_content = '#!/usr/bin/env bats\n\n\n'


    def setup(self):
        """Process arguments"""

        # Check the command-line options
        self.testfile     = self.get_parsed_argument(TESTFILE, "")
        self.output       = self.get_parsed_argument(OUTPUT, "")
        self.source       = self.get_parsed_argument(SOURCE, "")
        ## OLD: self.verbose      = self.has_parsed_option(VERBOSE)
        self.verbose      = (self.has_parsed_option(VERBOSE) or system.getenv("VERBOSE"))

        debug.trace(T7, (f'batspp - testfile: {self.testfile}, '
                         f'output: {self.output}, '
                         f'source: {self.source}, '
                         f'verbose: {self.verbose}'))


    def run_main_step(self):
        """Process main script"""

        # Check if is test of shell script file
        self.is_test_file = (TEST_FILE or self.testfile.endswith(BATSPP_EXTENSION))
        debug.trace(T7, f'batspp - {self.testfile} is a test file (not shell script): {self.is_test_file}')


        # Read file content
        self.file_content = system.read_file(self.testfile)
        if (self.is_test_file and PREPROCESS_BATSPP):
            self.file_content = preprocess_batspp(self.file_content)
        ## OLD
        ## if PARA_BLOCKS:
        ##     # Put spcial sentinel (U+FF) after perl-style paragraphs to facilitate isolating tests
        ##     # HACK: throws in extra newlines so that pattern matching works
        ##     debug.assertion("\xFF" not in self.file_content)
        ##     self.file_content = my_re.sub("\n\n", "\n\n\xFF", self.file_content)

        # Check if finish with newline
        if not self.file_content.endswith('\n\n'):
            self.file_content += '\n'


        self.__process_setup()
        self.__process_teardown()
        self.__process_tests()


        # Set Bats filename
        if self.output:
            batsfile = self.output
            if batsfile.endswith('/'):
                ## BAD: name = re.search(r"\/(\w+)\.", self.test_file).group(0)
                ## TODO: fixme (e.g., filename '##.batspp')-- ex via glue_helpers.remove_extension
                name = re.search(r"\/(\w+)\.", batsfile).group(0)
                batsfile += f'{name}.bats'
        else:
            batsfile = self.temp_file


        # Save Bats file
        gh.write_file(batsfile, self.bats_content)


        # Add execution permission to directly run the result test file
        if self.output:
            gh.run(f'chmod +x {batsfile}')


        # Run
        if not SKIP_BATS:
            debug.trace(T7, f'batspp - running test {self.temp_file}')
            bats_output = gh.run(f'bats {BATS_OPTIONS} {batsfile}')
            print(bats_output)
            debug.assertion(not my_re.search(r"^0 tests", bats_output, re.MULTILINE))


    def __process_setup(self):
        """Process tests setup"""
        debug.trace(T7, f'batspp - processing setup')


        # NOTE: files are loaded globally to
        #       to avoid problems with functions
        #       commands.


        # Make executables ./tests/../ visible to PATH
        #
        # This is usefull then the file is specially for testing.
        #
        # The structure should be:
        #  project
        #     ├ script.bash
        #  	  └ tests/test_script.batspp
        if not OMIT_PATH:
            self.bats_content += ('# Make executables ./tests/../ visible to PATH\n'
                                  f'PATH="{gh.dir_path(gh.real_path(self.testfile))}/../:$PATH"\n\n')

        # Enable aliases unless explicitly disabled
        enable_aliases = (not DISABLE_ALIASES)
        if not enable_aliases:
            all_content = (self.file_content + (system.read_file(self.source) if self.source else ""))
            debug.assertion(re.search(r"alias \w+\s*=", all_content))
        if enable_aliases:
            self.bats_content += ('# Enable aliases\n'
                                  'shopt -s expand_aliases\n\n')
            

        # Source Files
        if (self.source or (not self.is_test_file)):
            self.bats_content += ('# Source files\n')

            if not self.is_test_file:
                self.bats_content += f'source {gh.real_path(self.testfile)} || true\n'

            if self.source:
                self.bats_content += f'source {gh.real_path(self.source)} || true\n'

        self.bats_content += '\n\n'


    # pylint: disable=no-self-use
    def __process_teardown(self):
        """Process teardown"""
        debug.trace(T7, f'batspp - processing teardown')
        # WORK-IN-PROGRESS


    def __process_tests(self):
        """Process tests"""
        debug.trace(T7, f'batspp - processing tests')

        # Tests with simple indentation (i.e. #) are ignored on batspp files.
        # Tests with double indentation (i.e. ##) are ignored on shell scripts.
        global INDENT_PATTERN
        if self.is_test_file:
            ## Note: The {0} causes the following indent pattern to be ignored:
            ##   [^\w\$\(\n\{\}]
            INDENT_PATTERN += r'{0}'
            ## TODO: INDENT_PATTERN += "[^#]"
        else:
            INDENT_PATTERN += r'?'
        INDENT_PATTERN += r'\s*'

        command_tests  = CommandTests(verbose_debug=self.verbose)
        function_tests = FunctionTests(verbose_debug=self.verbose)

        # Do the test extraction, optionally removing regular content from bash scripts (i.e., extract comments and empty lines)
        file_content = self.file_content
        JUST_COMMENTS = system.getenv_bool("JUST_COMMENTS", (not self.is_test_file),
                                           "Strip non-comments from bash input")
        if JUST_COMMENTS:
            # Note: This simplifies pattern matching for text extraction, such as in helping
            # to avoid the expected output field from incorporating extraneous content.
            # TODO: allow for here-documents with comments and other special cases
            ## BAD: file_content = my_re.sub(r"^[^#][^\n]*\n", "\n", file_content, flags=re.MULTILINE)
            file_content = ""
            for line in self.file_content.splitlines():
                if not my_re.search("^[^#].*$", line):
                    file_content += line + "\n"
            debug.trace(T8, f"content after non-comment stripping:\n\t{file_content!r}")
            system.write_file(gh.form_path(TMP, "file_content.list"), file_content)
        #
        self.bats_content += command_tests.get_bats_tests(file_content)
        if not self.is_test_file:
            # TODO: sort combined set of tests based on file offset to make order more intuitive
            self.bats_content += function_tests.get_bats_tests(file_content)

#-------------------------------------------------------------------------------
        
# Global test number
_test_num = 0

class CustomTestsToBats:
    """Base class to extract and process tests"""

    def __init__(self, patterns, re_flags=0, verbose_debug=False):
        self._verbose       = verbose_debug
        self._re_flags      = re_flags
        assert(isinstance(patterns, list) and not isinstance(patterns, str))
        self._patterns      = None
        self._assert_equals = True
        self._setup_funct   = None
        ## OLD: self._test_id       = f'id{random.randint(1, 999999)}' # differentiate tests
        self._test_id       = self.next_id()
        self._indent_used   = None

        # Add optional header and trailer patterns
        self._patterns = patterns
        if MATCH_SENTINELS:
            # notes: This matches '# Start' as well as comments before title, which is
            # done in order to capture '# Setup' comments. Uses non-greedy matches to ignore test name.
            header  =  r'(?#header   )(?:# *Start[^\n]*\n(?:#[^\n]*\n)*?)'
            ## TODO: trailer =  r'(?#trailer  )(?:# *End[^\n]*\n)'
            trailer =  r'(?#trailer  )'
            self._patterns = [header] + patterns + [trailer]
        
    def next_id(self):
        """Return next ID for test"""
        global _test_num
        _test_num += 1
        test_id = str(random.randint(1, 999999) if RANDOM_ID else _test_num)
        debug.trace(T6, f"next_id() => {test_id}; self={self}")
        return test_id

    def _first_process(self, match):
        """First process after match, format matched results into [title, setup, actual, expected]"""

        # NOTE: this must be overriden.
        # TODO: raise NotImplementedError()
        # NOTE: the returned values must be like:
        ## OLD: result = ['title', 'setup', 'actual', 'expected']
        result = TestFieldTypes._fields

        debug.trace(T7, f'batspp (test {self._test_id}) - CustomTestsToBats._first_process({match}) => {result}')
        return result

    def _preprocess_field(self, field):
        """Preprocess match result FIELD"""

        # Remove comment indicators
        field = my_re.sub(r'^\# (Actual|Continuation|End|Setup|Start) *\n', '', field,
                          flags=re.MULTILINE|re.IGNORECASE)
            
        # Remove indent
        ## TODO?: if USE_INDENT_PATTERN and not Batspp.is_test_file:
        ## OLD:
        if not Batspp.is_test_file:
            field = my_re.sub(fr'^{self._indent_used}', '', field, flags=re.MULTILINE)

        return field

    def _preprocess_command(self, field):
        """Preprocess command FIELD"""
        field = self._preprocess_field(field)
        return field

    def _preprocess_output(self, field):
        """Preprocess output FIELD"""
        in_field = field
        field = self._preprocess_field(field)
        
        # Strip whitespaces
        # TODO: make this optional for expected output field
        field = field.strip()

        # Remove initial and trailing quotes
        field = my_re.sub(f'^(\"|\')(.*)(\"|\')$', r'\2', field)
        debug.trace(T8, f"_preprocess_output({in_field!r}) ==  {field!r}")
        return field

    def _common_process(self, test):
        """Common process for each field in test"""

        result = []

        # Get indent used
        self._indent_used = ''
        if my_re.match(INDENT_PATTERN, test.actual):
            self._indent_used = my_re.group(0)
            debug.trace(T7, f'batspp (test {self._test_id}) - indent founded: "{self._indent_used}"')

        # Preprocess command and output fields
        entire = test.entire
        title = test.title
        setup = self._preprocess_command(test.setup)
        actual = self._preprocess_command(test.actual)
        expected = self._preprocess_output(test.expected)

        result = TestFieldTypes(*[entire, title, setup, actual, expected])
        debug.trace(T7, f'batspp (test {self._test_id}) - _common_process({test}) => {result}')
        return result


    def _last_process(self, test):
        """Process test fields before convert into bats"""

        # NOTE: if this is overrided, the result must be:
        # result = ['title', 'setup', 'actual', 'expected']

        result = test

        debug.trace(T7, f'batspp (test {self._test_id}) - CustomTestsToBats._last_process({test}) => {result}')
        return result

    def _convert_to_bats(self, test):
        """Convert tests to bats format"""
        entire, title, setup, actual, expected = test
        debug.trace_expr(T6, entire, title, setup, actual, expected, prefix="_convert_to_bats: ", delim="\n")
        (actual, expected) = self.merge_continuation(actual, expected)

        # Process title
        # Note: The test label starts with a number and includes optional user name (e.g, test-1-hello-world).
        ## OLD: title = title if title else f'test {self._test_id}'
        title_prefix = f'test-{self._test_id}'
        if title:
            title = (title_prefix + "-" + title.replace(" ", "-"))
        else:
            title = title_prefix
        unspaced_title = my_re.sub(r'\s+', '-', title)

        # HACK: Extract setup command from entire match if '# Setup' indicator given
        # but the setup section is empty.
        if (EXTRACT_SETUP and (not setup) and my_re.search(r"# Setup\n\$([^\n]+\n)", entire,
                                                           flags=re.IGNORECASE|re.MULTILINE)):
            setup = my_re.group(1)
            debug.trace(T6, f"Using setup fallback code extracted from entire match: {setup}")
        # Similarly, get commands from entire section not present in setup nor actual
        if AUGMENT_COMMANDS:
            for command in re.findall(r"^\s*\$([^\n]+)\n", entire, flags=re.IGNORECASE|re.MULTILINE):
                if command.strip() not in (setup + actual + expected):
                    debug.trace(T6, f"Adding missing command to setup: {command}")
                    if setup and (not re.search(r"[;\}]\s*$", setup)):
                        setup += ";"
                    setup += ("\t" + command + "\n")
        debug.trace(T6, f"hacked_setup={setup!r}")

        # Process setup commands
        # Note: set COPY_DIR to copy files in current dir to temp. dir.
        ## TODO: temp_dir = Main.temp_base
        ## OLD:
        ## setup_text = (f'\ttestfolder=$(echo /tmp/{unspaced_title}-$$)\n'
        ##               f'\tmkdir $testfolder && cd $testfolder\n')
        gh.full_mkdir(TEMP_DIR)
        setup_text = (
                      ## TEST: ("" if not GLOBAL_SETUP.strip() else ("\t" + GLOBAL_SETUP + ";\n")) +
                      f'\ttestfolder="{TEMP_DIR}/{unspaced_title}"\n' +
                      f'\tmkdir --parents "$testfolder"\n' +
                      (f'\tcommand cp ./*.* "$testfolder"\n' if COPY_DIR else '') +
                      # note: warning added for sake of shellcheck
                      f'\tcd "$testfolder" || echo Warning: Unable to "cd $testfolder"\n')
        ## BAD: (splitlines doesn't account for line continuations):
        ## for command in setup.splitlines():
        ##    if command:
        ##        OLD: command     = my_re.sub(r'(^\$\s+|\n+$)', '', command, flags=re.MULTILINE)
        ##        debug.assertion("\n" not in command)
        ##        command     = my_re.sub(r'^\$\s+', '', command, re.MULTILINE)
        ##        setup_text += f'\t{command}\n'
        setup_sans_prompt = my_re.sub(r'^\s*\$', '\t', setup, flags=re.MULTILINE)
        ## OLD
        ## HACK: fixup lingering $
        ## setup_sans_prompt = setup_sans_prompt.replace("\n$", "\n")
        setup_text += setup_sans_prompt + "\n"
        debug.trace_expr(T6, setup_text)

        # actual and expected
        actual_label   = 'actual'
        expected_label = 'expected' if self._assert_equals else 'not_expected'
        setup_label   = 'setup'


        ## OLD: (moved later)
        ## Process debug
        ##
        ## verbose_print = '| hexview.perl' if  self._verbose else ''
        ## debug_text = (f'\techo "==========" ${actual!r} "=========="\n'
        ##               f'\techo ${actual!r} {verbose_print}\n'
        ##               f'\techo "=========" ${expected!r} "========="\n'
        ##               f'\techo ${expected!r} {verbose_print}\n'
        ##               '\techo "============================"\n')


        # Process assertion
        assertion_text = "==" if self._assert_equals else "!="


        # Process functions
        actual_function   = f'{unspaced_title}-{actual_label}'
        expected_function = f'{unspaced_title}-{expected_label}'
        setup_function    = None
        functions_text    = ''
        functions_text   += self._get_bash_function(actual_function, actual)
        ## BAD: functions_text   += self._get_bash_function(expected_function, f'echo -e {repr(expected)}')
        # Note: to minimize issues with bash syntax, a bash here-document is used (e.g., <<END\n...\nEND\n).
        # TOOO?: Use <<- variant so that leading tabs are ingored.
        # TODO: use an external file (as the @here might fail if the example uses << as well)
        expected_output = ('\tcat <<END_EXPECTED\n' +
                           (expected + "\n") +
                           'END_EXPECTED')
        functions_text   += self._get_bash_function(expected_function, expected_output,
                                                    output=True)

        # Add special hooks for when '# Setup'or '# Continuation'  specified
        # HACK: Updates instance state to process such indicator comments (to avoid regex complication)
        has_setup_comment = re.search("# Setup", entire, re.IGNORECASE)
        if has_setup_comment:
            setup_function   = f'{unspaced_title}-{setup_label}'
            self._setup_funct = setup_function
            functions_text += self._get_bash_function(setup_function, setup_text)
        has_continuation_comment = re.search("# Continuation", entire, re.IGNORECASE)
        use_setup_function = (has_setup_comment or has_continuation_comment)
        setup_call = ""
        if use_setup_function:
            debug.trace(T6, f"Using separate setup function {self._setup_funct}")
            if not self._setup_funct:
                system.print_error(f"Error: No setup function defined for test {unspaced_title}")
            else:
                setup_call = ("\t" + self._setup_funct + ";\n")
        else:
            ## TODO: self._setup_funct = None
            pass

        ## Process debug
        debug_text = ""
        if not OMIT_TRACE:
            verbose_print = '| hexview.perl' if  self._verbose else ''
            debug_text = (f'\techo "==========" ${actual!r} "=========="\n'
                          f'\t{actual_function} {verbose_print}\n'
                          f'\techo "=========" ${expected!r} "========="\n'
                          f'\t{expected_function} {verbose_print}\n'
                          '\techo "============================"\n')

        # Construct bats tests
        misc_code = ""
        if not OMIT_MISC:
            misc_code = (
                f'\t# ???: {actual!r}=$({actual_function})\n' +         # TODO: fix (Bruno, what is for?)
                f'\t# ???: {expected!r}=$({expected_function})\n')      # TODO: fix (ditto)
        result = (f'@test "{title}" {{\n' +
                  (f'{setup_text}' if not use_setup_function else setup_call) +
                  f'{debug_text}' +
                  misc_code + 
                  ## BAD: f'\t[ ${actual!r} {assertion_text} ${expected!r} ]\n' +
                  f'\t[ "$({actual_function})" {assertion_text} "$({expected_function})" ]\n' +
                  f'}}\n\n' +
                  f'{functions_text}\n')

        debug.trace(T7, f'batspp (test {self._test_id}) - _convert_to_bats({test}) =>\n{result}')
        return result

    def _get_bash_function(self, name, content, output=False):
        """Return bash function with NAME and code CONTENT, optionally for expected OUTPUT"""
        if not output:
            # Strip comments
            content = my_re.sub(r"^\s*\#[^\n]+\n", "", content, flags=re.MULTILINE)
            # Remove prompts
            content = my_re.sub(r"^\s*\$ ", "", content, flags=re.MULTILINE)
        # Make sure indented
        if not content.startswith("\t"):
            content = my_re.sub(r"^([^\t])", r"\t\1", content, flags=re.MULTILINE)
        result = (f'function {name} () {{\n' +
                  '\t# no-op in case content just a comment\n' +
                  '\ttrue\n' +
                  '\n' +
                  f'{content}\n' +
                  '}\n\n')
        debug.trace(T7, f'batspp (test {self._test_id}) - get_bash_function(name={name}, content={content}) => {result}')
        return result

    def trace_pattern_match(self, text):
        """Traces out the matching for TEXT in stages, using self._patterns one at a time. 
        This uses a greedy approximation to the regular regex matching with concatenation.
        Note: Intended just as quick-n-dirty way to debug regex evaluation
        """
        debug.trace(T6, f"trace_pattern_match({gh.elide(text, 512)!r})")
        for p, pattern in enumerate(self._patterns):
            debug.trace(T7, f"\tp{p + 1}: {pattern}")
        # note: temporarily raises level of overly verbose regex matching
        save_my_re_TRACE_LEVEL = my_re.TRACE_LEVEL
        my_re.TRACE_LEVEL = T9
        save_INDENT0 = debug.INDENT0

        # Initialize
        start = 0
        p = 0
        num_patterns = len(self._patterns)
        matches = []

        # Do searches incrementally
        found = False
        while ((p < num_patterns) and (not found)):
            next_p = (p + 1)
            debug.INDENT0 = ("    " * (p + 1))

            # Do next search
            # Note: If text consumed, search is over dummy newline.
            ## OLD: pattern_spec = f"{p + 1} (of {len(self._patterns)})"
            pattern_spec = f"{p + 1}"
            debug.trace(T6, f"Checking pattern {pattern_spec} at offset {start}; {gh.elide(text[start:], 10)!r}")
            over_actual_text = (len(text[start:]) > 0)
            remainder = (text[start:] if over_actual_text else "\n")
            searc_fn = (my_re.match if (p > 0) else my_re.search)
            found = searc_fn(self._patterns[p], remainder, flags=self._re_flags)
            if not found:
                break
            start += my_re.end()
            match_text = my_re.group(0)
            matches.append(match_text if over_actual_text else f"{match_text}???")
            debug.trace(T6, f"Match for pattern {pattern_spec} at {my_re.span()}: {gh.elide(my_re.group(0), 80)!r}")
            found = (next_p == num_patterns)

            # Advance to next pattern
            p += 1
        debug.INDENT0 = ""

        # Summarize
        debug.trace_expr(T6, found)
        if found:
            debug.trace(T6, "Matching text:")
            for m in range(len(matches)):
                debug.trace(T6, f"\tp{m + 1}: {matches[m]!r}")

        # Reset temporary changes
        my_re.TRACE_LEVEL = save_my_re_TRACE_LEVEL
        debug.INDENT0 = save_INDENT0
        return

    def normalize_block(self, text):
        """Normalize TEXT for matching (e.g., adding Start/End sentinels)"""
        debug.trace(T9, f"normalize_block(_); self={self}")
        in_text = text
        debug.assertion(PARA_BLOCKS)
        if MATCH_SENTINELS:
            if not my_re.search("# Start", text):
                text = "# Start\n" + text.lstrip("\n")
            ## TODO:
            ## if not my_re.search("# End", text):
            ##     text = text.rstrip("\n") + "\n# End\n"
        debug.trace(T8, f"normalize_block({in_text!r}) => {text!r}")
        return text
    
    def get_bats_tests(self, text):
        """Returns BATS tests"""

        debug.trace(T7, f'batspp - pattern used ({self._re_flags}): {self._patterns}')

        bats_tests = ''

        ## TODO:
        ## NORMALIZE_WHITESPACE = system.getenv_bool("NORMALIZE_WHITESPACE", False,
        ##                                           "Convert non-newline whitespace to space")
        ## if NORMALIZE_WHITESPACE:
        ##     text = my_re.sub(r"\s(\S)", r" \1", text)
        ##     debug.trace(T9, f"after whitespace normalization: text={text!r}")
        
        # Find all patterns in the text. In special perl-style paragraph mode, the tests
        # must reside within blocks delimited by double newlines (i.e., \n\n).
        all_matches = []
        pattern = "(" + "".join(self._patterns) + ")"
        debug.trace_expr(T7, pattern)
        if not PARA_BLOCKS:
            all_matches = re.findall(pattern, text, flags=self._re_flags)
            if (TRACE_MATCHING and ((len(all_matches) == 0) or debug.debugging(T6))):
                self.trace_pattern_match(text)
        else:
            debug.assertion("\xFF" not in text)
            text = my_re.sub(r"(\n\n+)", "\\1\xFF", text)
            for sub_text in text.split("\xFF"):
                sub_text = self.normalize_block(sub_text)
                sub_matches = re.findall(pattern, sub_text, flags=self._re_flags)
                if (TRACE_MATCHING and ((len(all_matches) == 0) or debug.debugging(T6))):
                    self.trace_pattern_match(sub_text)
                debug.trace_expr(T6, sub_text, len(sub_matches))
                all_matches += sub_matches

        # Process each match
        for match in all_matches:
            debug.trace(T7, f'batspp (test {self._test_id}) - processing match: {match}')
            ## OLD: debug.assertion(len(match) >= 4, 'Insufficient groups, you should review the used regex pattern.')
            debug.assertion(len(match) == 5, 'Incorrect number of groups (see TextTypes), you should review the used regex pattern.')

            test = self._first_process(match)
            debug.assertion(len(test) == 5, f'Incorrect number of fields, every test should be {TestFieldTypes._fields}')
            # Add global setup section to directly BATS output mostly as is (except for prompt removal)
            # pylint: disable=no-member
            if re.search(r"^\s*# Global Setup", test.entire, flags=(re.MULTILINE | re.IGNORECASE)):
                global_setup = my_re.sub(r'^\s*\$\s*', '', test.entire, flags=re.MULTILINE)
                debug.trace_expr(T6, global_setup)
                bats_tests += global_setup + "\n"
                continue
            
            test = self._common_process(test)
            test = self._last_process(test)
            bats_tests += self._convert_to_bats(test)

            ## OLD: self._test_id = f'id{random.randint(1, 999999)}'
            self._test_id = self.next_id()

        return bats_tests

    def merge_continuation(self, actual, expected):
        """Merge EXPECTED content to ACTUAL if likely contintuation lines"""
        debug.trace(T7, f"merge_continuation(_c, _s): len(a)={len(actual)}; len(e)={len(expected)}; self={self}")
        # notes: accounts for function definitions and line continuations.
        # also, experimental code, so enabled with MERGE_CONTINUATION environment setting.
        if (MERGE_CONTINUATION and actual and expected):
            while True:
                new_actual, new_expected = merge_continuation(actual, expected)
                if (new_actual == actual):
                    debug.assertion(new_expected == expected)
                    break
                if ((len(new_actual) + len(new_expected)) != (len(actual) + len(expected))):
                    debug.trace(T6, "Unexpected merge results")
                    break
                actual = new_actual
                expected = new_expected
        return (actual, expected)


class CommandTests(CustomTestsToBats):
    """Extract and process specific command tests"""
    ## TODO: add examples to help clarify the regex

    def __init__(self, verbose_debug=False):
        flags = re.DOTALL | re.MULTILINE | re.IGNORECASE

        # Notes:
        # - Non-capturing groups '(?:regex)' used for ignore certain groupings.
        # - INDENT_PATTERN='^[^\w\$\(\n\{\}]{0}\s*' for batspp file (e.g., from Jupyter).
        # - The optional setup is for commands issued but output not tested.
        # - Example 1:
        #     # Test fu                                              # title
        #     $ cd /tmp                                              # setup
        #     $ pwd                                                  # actual
        #     /tmp                                                   # expected
        # - Example 2:
        #     $ uname                                                # actual
        #     Linux                                                  # expected
        # - The MATCH_SENTINELS option add patterns to capture more context.
        debug.trace_expr(T8, self, INDENT_PATTERN)

        if USE_INDENT_PATTERN:
            patterns = [
                fr'(?:{INDENT_PATTERN} *# Test\s*([^\n]*)\s*\n)?+', # optional test title
                fr'((?:{INDENT_PATTERN}\$\s+[^\n]+\n)*)',           # optional setup
                fr'({INDENT_PATTERN}\$\s+[^\n]+)\n',                # command line [actual]
                r'(.+?)\n',                                         # expected output; non-greedy
                fr'{INDENT_PATTERN}$']                              # end test
        else:
            patterns = [
                r'(?#title    )(?:^ *\# *Test +([^\n]*)\n)?',      # optional test title
                # optional setup code w/ explicit start ('# setup') and end ('# actual') indicators
                r'(?#setup    )((?:^ *\# *Setup\n)'        \
                              r'(?:^ *\#? *\$ +[^\n]+\n)*' \
                              r'(?:^ *\# *Actual\n))?',
                r'(?#actual   )(^ *\#? *\$ +[^\n]+\n)',            # command line [actual]
                r'(?#expected )(.*)',                              # expected output
                r'(?#end      )^ *\#?\s*\n']                       # end test (blank line)

        super().__init__(patterns, re_flags=flags, verbose_debug=verbose_debug)


    def _first_process(self, match):
        """First process after match, format matched results into [entire, title, setup, actual, expected]"""

        #         title     setup     actual    expected

        assert (len(match) == 5)
        ## OLD: result = [match[1], match[2], match[4], match[5]]
        entire, title, setup, actual, expected = match
        actual, expected = self.merge_continuation(actual, expected)

        result = TestFieldTypes(*[entire, title, setup, actual, expected])
                
        debug.trace(T7, f'batspp (test {self._test_id}) - CommandTests._first_process({match}) => {result}')
        return result


    def _last_process(self, test):
        """Process test fields before convert into bats"""

        entire, title, setup, actual, expected = test
        actual, expected = self.merge_continuation(actual, expected)
        
        # Remove $
        # TODO?: flags=re.MULTILINE
        ## OLD: actual = my_re.sub(r'^\$\s*', '', actual)
        actual = my_re.sub(r'^\$ ', '', actual)
        setup = my_re.sub(r'^\$ ', '', setup, flags=re.MULTILINE)

        ## OLD: result = title, setup, actual, expected
        result = TestFieldTypes(*[entire, title, setup, actual, expected])
        debug.trace(T7, f'batspp (test {self._test_id}) - CommandTests._last_process({test}) => {result}')
        return result


class FunctionTests(CustomTestsToBats):
    """Extract and process specific function tests"""
    ## TODO: add examples (e.g., to help distinguish regex from one for CommandTests)


    # Class globals
    assert_eq     = r'=>'
    assert_ne     = r'=\/>'


    def __init__(self, verbose_debug=False):

        flags = re.MULTILINE | re.IGNORECASE

        # Notes:
        # - Non-capturing groups '(?:regex)' used for to ignore certain groupings
        # - INDENT_PATTERN='^[^\w\$\(\n\{\}]{0}\s*' for batspp file (e.g., from Jupyter)
        # - Example 1:
        #     # Test fibonacci function                             # title
        #     fibonacci 9 => "0 1 1 2 3 5 8 13 21 34"               # actual assertion-EQ expected
        # - Example 2:
        #     fibonacci 3 =/> "8 2 45 34 3 5"                       # actual assertion-NE expected
        # - Entire match included as well (see get_bats_tests).
        # - MATCH_SENTINELS not needed as => and =/=> constrain the search.
        debug.trace_expr(T8, self, INDENT_PATTERN)

        if USE_INDENT_PATTERN:
            patterns = [
                fr'(?:{INDENT_PATTERN} *# Test\s*([^\n]*)\s*\n)?',   # optional test title
                fr'({INDENT_PATTERN}[^\n]+\w[^\n]+)',                # functions + args line [actual]; '.+?' ???
                fr'\s+({self.assert_eq}|{self.assert_ne})\s+',       # assertion
                r'((?:.|\n)*)',                                      # expected output
                fr'\n{INDENT_PATTERN}$']                             # blank indent (end test)
        else:
            EQ = self.assert_eq
            NE = self.assert_ne
            patterns = [
                r'(?#title    )(?:# Test\s*([^\n]*)\s*\n)?',         # optional test title
                # optional setup code w/ explicit start ('# setup') and end ('# actual') indicators
                r'(?#setup    )((?:^ *\# *Setup\n)'        \
                              r'(?:^ *\#? *\$ +[^\n]+\n)*' \
                              r'(?:^ *\# *Actual\n))?',
                # functions + args line [actual] with assertion
                r'(?#actual   )([^\n]+\w[^\n]+'            \
                              fr'\s+(?:{EQ}|{NE}))\s+',              # assertion
                r'(?#expected )([^\n]+)']                            # expected output

        super().__init__(patterns, re_flags=flags, verbose_debug=verbose_debug)


    def _first_process(self, match):
        """First process after match, format matched results into [entire, title, setup, actual, expected]"""

        ## OLD: _, title, actual, assertion, expected, _ = match
        entire, title, setup, actual, expected = match

        # Check assertion
        # HACK: assertion put into instance
        # TODO: use OO apprach (e.g., FunctionResult w/ assertion field versus CommandResult with setup)
        ## OLD:
        ## if assertion == self.assert_eq:
        ##     self._assert_equals = True
        ## elif assertion == self.assert_ne:
        ##     self._assert_equals = False
        ## else:
        ##     self._assert_equals = None
        if my_re.search(fr'(.*)({self.assert_eq}|{self.assert_ne}).*$', actual):
            actual = my_re.group(1)
            self._assert_equals = (my_re.group(2) == self.assert_eq)

        # Format values
        # Note: uses empty setup
        ## OLD: result = [entire, title, '', actual, expected]
        result = TestFieldTypes(*[entire, title, setup, actual, expected])

        debug.trace(T7, f'batspp (test {self._test_id}) - _first_process({match}) => {result}')
        return result


    def _last_process(self, test):
        """Process test fields before convert into bats"""

        entire, title, setup, actual, expected = test

        # Disable alias adding '//' to functions
        ## TODO: example
        ## BAD: actual = '\\' + actual

        ## OLD: result = [entire, title, setup, actual, expected]
        result = TestFieldTypes(*[entire, title, setup, actual, expected])
        debug.trace(T7, f'batspp (test {self._test_id}) - _last_process({test}) => {result}')
        return result

#-------------------------------------------------------------------------------

if __name__ == '__main__':
    app = Batspp(description          = __doc__,
                 ## TODO: use_temp_base_dir=True,
                 positional_arguments = [(TESTFILE, 'test file path')],
                 boolean_options      = [(VERBOSE,  'show verbose debug')],
                 text_options         = [(OUTPUT,   'target output .bats filepath'),
                                         (SOURCE,   'file to be sourced')],
                 manual_input         = True)
    app.run()



#! /usr/bin/env python
#
# Trains text categorizer using Scikit-Learn. See
#    http://scikit-learn.org/stable/tutorial/text_analytics/working_with_text_data.html
#
# Notes:
# - This is a simple wrapper around the code in text_categorizer.py, created
#   to clarify how training is done.
#
# TODO:
# - *** Use sklearn for spliting training data.
# - ** Rework via main.py.
# - Rename to something like apply_text_categorizer.py, as now supports
#   usage with a pre-trained model.
#
#

"""Trains text categorization"""

# Standard packages
import math
import sys

# Local packages
from mezcla import debug
from mezcla import glue_helpers as gh
from mezcla.main import Main
from mezcla import system
from mezcla.text_categorizer import TextCategorizer

# Constants
SHOW_REPORT = system.getenv_bool("SHOW_REPORT", False,
                                 "Show categorization report from Sklearn")
TEST_PERCENT = system.getenv_value("TEST_PERCENT", None,
                                   "Percent of testing data to derive from training")

def usage():
    """Show command-line usage"""
    # TODO: remove path from script filename
    # TODO2: use stdout (as with argparse-based usage)
    script = (__file__ or "n/a")
    print("Usage: {scr} training-file model-file [testing]".format(scr=script))
    print("")
    print("Notes:")
    print("- Use - to indicate the file is not needed (e.g., existing training model).")
    print("- You need to supply either training file or model file.")
    print("- The testing file is optional when training.")
    print("- Currently, only tab-separated value (TSV) format is accepted:")
    print("    <label>\t<text>")
    return


def main(args=None):
    """Entry point for script"""
    debug.trace_fmtd(4, "main(): args={a}", a=args)

    # Check command line arguments
    if args is None:
        args = sys.argv
        debug.trace_fmtd(4, "len(args)={l}", l=len(args))
        if (len(args) == 1) and system.getenv_text("ARGS"):
            args += system.getenv_text("ARGS").split()
            debug.trace_fmtd(4, "len(args)={l}; args={a}", l=len(args), a=args)
        if len(args) <= 2:
            usage()
            return
    training_filename = args[1]
    model_filename = args[2]
    testing_filename = None
    if (len(args) > 3):
        testing_filename = args[3]

    # Note: This gets some settings from main, but it is not used directly.
    app = Main(skip_args=True)
        
    # Optionally infer testing data from training data
    # TODO: support randomization
    if TEST_PERCENT:
        debug.assertion((testing_filename is None) or (testing_filename == "-"))
        full_training_filename = training_filename
        total_num_lines = system.to_int(gh.run(f"wc -l < {full_training_filename}"))
        num_testing_lines = int(math.ceil((system.to_float(TEST_PERCENT) / 100.0) * (total_num_lines - 1)))
        num_training_lines = (total_num_lines - num_testing_lines)
        debug.trace_expr(5, total_num_lines, num_testing_lines, num_training_lines)
        
        # Split the training file into training proper and test
        # note: the tail --lines=+2 option ignores the header line
        training_filename = app.temp_base + "-train.tsv"
        debug.assertion(not testing_filename)
        testing_filename = app.temp_base + "-test.tsv"
        gh.issue(f"head --lines=1 < {full_training_filename} > {training_filename}")
        gh.issue(f"tail --lines=+2 < {full_training_filename} | head --lines={num_training_lines} >> {training_filename}")
        testing_filename = app.temp_base + "test.tsv"
        gh.issue(f"head --lines=1 < {full_training_filename} > {testing_filename}")
        gh.issue(f"tail --lines=+2 < {full_training_filename} | tail --lines={num_testing_lines} >> {testing_filename}")
        gh.run(f"wc -l {full_training_filename} {app.temp_base}*")
        
    # Train text categorizer and save model to specified file
    text_cat = TextCategorizer()
    new_model = False
    accuracy = None
    try:
        if training_filename and (training_filename != "-"):
            text_cat.train(training_filename)
            new_model = True
        if model_filename and (model_filename != "-"):
            if new_model:
                text_cat.save(model_filename)
            else:
                text_cat.load(model_filename)
        if testing_filename and (testing_filename != "-"):
            accuracy = text_cat.test(testing_filename, report=SHOW_REPORT)
            print("Accuracy over {f}: {acc}".format(acc=accuracy, f=testing_filename))
            debug.trace_fmt(4, "Tested using classifer {clf}",
                            clf=text_cat.classifier.named_steps['clf'].__class__)
    except:
        system.print_exception_info("categorization")
        debug.raise_exception(6)

    # Show usage if nothing actually done (e.g., due to too many -'s for filenames)
    ## OLD: if (not (new_model or accuracy)):
    if (not (new_model or (accuracy is not None))):
        usage()
              
    return

#------------------------------------------------------------------------

if __name__ == '__main__':
    main()



#!/usr/bin/env python3

"""Configuration settings"""

# Installed modules
## TEST: from cachetools import LRUCache, cached
from mezcla import debug
from mezcla import system

# Constants
TL = debug.TL
BASE_DEBUG_LEVEL = system.getenv_int(
    "TFIDF_DEBUG_LEVEL", 6,
    description="Base level for TF/IDF tracing--set to {TL.USUAL} for isolated debugging")
BDL = BASE_DEBUG_LEVEL

#-------------------------------------------------------------------------------

def main():
    """Entry point for script"""
    debug.trace(BDL + -2, "Warning: Unimplemented main")

#-------------------------------------------------------------------------------
    
if __name__ == '__main__':
    system.print_stderr(f"Warning: {__file__} is not intended to be run standalone.")
    main()



#! /usr/bin/env python
#
# Miscellaneous tests not tied to particular module.
#

"""Miscellaneous/non-module tests"""

# Standard packages
## TODO: from collections import defaultdict

# Installed packages
import pytest

# Local packages
from mezcla import debug
from mezcla import system
from mezcla.unittest_wrapper import TestWrapper
from mezcla.unittest_wrapper import trap_exception
from mezcla.tfidf.corpus import Corpus
from mezcla.tfidf.dockeyword import DocKeyword

# Constants
DOCUMENT_DATA = [
    "0",
    "1 2 3 4 5",
    "2 4",
    "3 4 5",
    ""
    ]
# note: order is lexicographic in case of ties
TOP_BIGRAMS =  ["",  "1 2",  "2 4",  "3 4",  ""]
TOP_UNIGRAMS = ["0",  "1",    "2",    "3",    ""]

# Environment settings
HOME = system.getenv_text(
    "HOME", "~",
    description="Home directory for user")
USING_GITHUB_RUNNER = ("/home/runner" in HOME)

class TestMisc(TestWrapper):
    """Class for test case definitions"""
    script_module= None
    use_temp_base_dir = True

    @pytest.fixture(autouse=True)
    def capsys(self, capsys):
        """Gets capsys"""
        self.capsys = capsys

    @pytest.mark.xfail                   # TODO: remove xfail
    @trap_exception
    def test_01_simple_tfidf(self):
        """Verify top term for each document by TF/IDF using unigrams and bigrams"""
        debug.trace(4, "test_01_simple_tfidf()")
        uni_corpus = Corpus(min_ngram_size=1, max_ngram_size=1)
        bi_corpus = Corpus(min_ngram_size=2, max_ngram_size=2)

        # Add text to collections
        for d, doc_text in enumerate(DOCUMENT_DATA):
            uni_corpus[d + 1] = doc_text
            bi_corpus[d + 1] = doc_text

        # Make sure top term is expected (n.b., sorted in case of ties)
        debug.assertion(len(DOCUMENT_DATA) == len(uni_corpus.keys()))
        for d, docid in enumerate(uni_corpus.keys()):
            #
            def is_top_term(top_term, corpus_info):
                """Whether TOP_TERM is top for CORPUS_INFO document D"""
                ## OLD: corpus_terms = [term for (term, _score) in list(corpus_info)]
                corpus_terms = [keyword.ngram for keyword in list(corpus_info)]
                ok = False
                if not top_term:
                    ok = not corpus_terms
                else:
                    ok = (top_term == sorted(corpus_terms)[0])
                debug.trace_expr(5, top_term, corpus_terms)
                return ok
            #
            self.do_assert(is_top_term(TOP_UNIGRAMS[d], uni_corpus.get_keywords(docid)))
            self.do_assert(is_top_term(TOP_BIGRAMS[d], bi_corpus.get_keywords(docid)))
        debug.trace(5, "out test_01_simple_tfidf()")

    @pytest.mark.skipif(not USING_GITHUB_RUNNER,
                        reason="The following is only for the Github runner VM")
    def test_02_no_dockeyword_tracing(self):
        """Make sure no debug calls in DocKeyword class
           Note: This is intended as a sanity check to ensure the debug calls aren't
           checked in by accident. That it, the '## DEBUG:' prefix should only be
           removed in a local version."""
        # Set the highest-possible debugging level
        ## TODO2: use monkey patch
        save_debug_level = debug.get_level()
        debug.set_level(system.MAX_INT)
        # note: make sure stderr empty by capturing it before initializer (DocKeyword)
        pre_init_stderr_text = self.capsys.readouterr().err

        # Perform the test proper
        dummy_word = "dummy"
        dk = DocKeyword(dummy_word)
        init_stderr_text = self.capsys.readouterr().err
        self.do_assert(not init_stderr_text)
        self.do_assert(dk.text == dummy_word)

        # Post-test sanity checks
        debug.assertion(isinstance(pre_init_stderr_text, str))
        debug.assertion(isinstance(init_stderr_text, str))
        debug.set_level(save_debug_level)
        debug.trace_expr(5, pre_init_stderr_text, init_stderr_text, dk)



#! /usr/bin/env python
#
# TLDR: Attempt at getting maldito pytest not to swallow exceptions.
#
# Note:
# - *** This doesn't work: instead use trap_exception decorator in unittest_wrapper.py.
# - Based on https://stackoverflow.com/questions/62419998/how-can-i-get-pytest-to-not-catch-exceptions.
# - See https://docs.pytest.org/en/7.1.x/reference/reference.html.
#

"""Configuration so that pytest doesn't swallow exceptions"""

# Standard packages
## OLD: import os

# Installed packages
## TODO: import pytest

# Local packages
from mezcla import debug
from mezcla import system

# Have pytest enable exceptions
ENABLE_PYTEST_RAISE = system.getenv_bool('_PYTEST_RAISE', False,
                                         desc="Raise exceptions in test code")
if ENABLE_PYTEST_RAISE:

    ## OLD
    ## @pytest.hookimpl(tryfirst=True)
    ## def pytest_exception_interact(call):
    ##     debug.trace(5, "pytest_exception_interact({call})")
    ##     raise call.excinfo.value
    ##
    ## @pytest.hookimpl(tryfirst=True)
    ## def pytest_internalerror(excinfo):
    ##     debug.trace(5, "pytest_internalerror({excinfo})")
    ##     raise excinfo.value
    
    def pytest_exception_interact(node, call, report):
        """Called when an exception was raised which can potentially be interactively handled"""
        debug.trace(5, f"pytest_exception_interact{((node, call, report))}")
        raise call.excinfo.value

    def pytest_internalerror(excrepr, excinfo):
        """Return True to suppress the fallback handling of printing an INTERNALERROR message directly to sys.stderr"""
        debug.trace(5, f"pytest_internalerror{(excrepr, excinfo)}")
        raise excinfo.value
    

#------------------------------------------------------------------------

if __name__ == '__main__':
    debug.trace_current_context()
    system.print_stderr("Error: Not intended to be invoked directly")



#! /usr/bin/env python
#
# extract_document_text.py: extract text from documents of various types
#
# NOTE:
# - Mostly a wrapper around functionality in textract package.
# - By default a suffix is added to the entire filename (e.g., fubar.docx.txt.1apr18),
#   but an affix can be used instead (e.g., fubar.docx.1apr18.txt).
#
# TODO:
# - Use file date for default date affix.
# - Handle UTF8.
#
#

"""Extract text from common document types"""

# Standard packages
import sys
import datetime

# Installed packages
import textract

# Local packages
from mezcla import system
from mezcla import debug
from mezcla import glue_helpers as gh


# Determine environment-based options
STDOUT = system.getenv_boolean("STDOUT", False)
## TODO: USE_TODAY = system.getenv_boolean("USE_TODAY", False)
## TODO: use new misc_util.get_date_ddmmmyy()
TODAY_DDMMMYY = datetime.date.today().strftime("%d%b%y").lower()
ADD_DATE = system.getenv_boolean("ADD_DATE", False)
DEFAULT_SUFFIX = "" if (STDOUT or not ADD_DATE) else TODAY_DDMMMYY
HTML = system.getenv_bool("HTML", False, desc="Assume HTML input")
DOC = system.getenv_bool("DOC", False, desc="Assume MS document input")
DEFAULT_IN_TYPE = ("html" if HTML else "doc" if DOC else "txt")
IN_TYPE = system.getenv_text("IN_TYPE", DEFAULT_IN_TYPE,
                             desc="Input file type--ext w/o period")
OUT_TYPE = system.getenv_text("OUT_TYPE", "txt",
                              desc="Input file type--ext w/o period")
EXT = system.getenv_text("EXT", f".{IN_TYPE}")
IN_EXT = EXT
SUFFIX = system.getenv_text("SUFFIX", DEFAULT_SUFFIX)
USE_AFFIX = system.getenv_boolean("USE_AFFIX", False)
FORCE = system.getenv_boolean("FORCE", False)
EXTENSION = system.getenv_value("EXTENSION", f".{OUT_TYPE}",
                                "Extension of file for conversion")
OUT_EXT = EXTENSION

def show_usage_and_quit():
    """Show command-line usage for script and then exit"""
    ## OLD:
    ## usage = """
## Usage: {prog} [options] [input-dir] [output-file]

    usage = """
Usage: {prog} [options] [input-file] ...

Options: [--help]

Notes:
- Converts documents to text
- Use following environment options:
  STDOUT: Send all output to standard output.
  IN_EXT: File type for unnamed input (e.g., .html)
  OUT_EXT: File extension for output text file (e.g., .txt)
  SUFFIX: New suffix for output text file (e.g., [.txt].copy).
  ADD_DATE: Add date as suffix or affix (e.g., .txt.17mar18).
  USE_AFFIX: Put suffix before file extension (e.g., .copy[.txt]).
"""
    print(usage.format(prog=sys.argv[0]))
    sys.exit()


def document_to_text(doc_filename):
    """Returns text version of document FILENAME of unspecified type"""
    debug.trace(4, f"document_to_text({doc_filename})")
    debug.trace_expr(4, IN_EXT)
    text = ""
    OK = False
    try:
        text = textract.process(doc_filename).decode("UTF-8")
        OK = True
    except ImportError:
        debug.trace_fmtd(3, "FYI: import error converting file {f}: {e}",
                         f=doc_filename, e=sys.exc_info())

    if not OK:
        try:
            ## OLD: text = system.from_utf8(textract.process(doc_filename))
            text = textract.process(doc_filename,
                                    extension=IN_EXT
                                    ).decode("UTF-8")
        except:
            debug.trace_fmtd(3, "Warning: problem converting document file {f}: {e}",
                             f=doc_filename, e=sys.exc_info())
    debug.trace(4, f"document_to_text() => {gh.elide(text)!r}")
    return text


def main():
    """Entry point for script"""
    args = sys.argv[1:]
    debug.trace_fmtd(5, "main(): args={a}", a=args)
    affix = ""
    suffix = ""
    if SUFFIX:
        if USE_AFFIX:
            affix = ("." + SUFFIX)
        else:
            suffix = ("." + SUFFIX)
    debug.trace_fmtd(4, "affix={a} suffix={s}", a=affix, s=suffix)
    use_stdout = STDOUT

    # Parse command-line arguments
    i = 0
    if (not args):
        show_usage_and_quit()
    while (args[i].startswith("--")):
        option = args[i]
        if (option == "--help"):
            show_usage_and_quit()
        else:
            sys.stderr.write("Error: unknown option '{o}'\n".format(o=option))
            show_usage_and_quit()
        i += 1

    # Process each of the arguments
    for filename in args:
        # Use stdio if no filename specified (i.e., -)
        if (filename == "-"):
            filename = gh.get_temp_file() + IN_EXT
            system.write_file(filename, system.read_all_stdin())
            use_stdout = True

        # Warning that HTML not well supported
        if (not HTML and filename.endswith(".html")):
            system.print_error("Warning: HTML not well supported; use html_utils.html_to_text instead")
            
        # Read in file contents
        text = document_to_text(filename)

        # Write output, using text extension unless stdout
        ## TODO: file_date = datetime.datetime.fromtimestamp(os.path.getmtime(filename))
        if use_stdout:
            sys.stdout.write(system.to_utf8(text) + "\n")
        else:
            new_filename = system.remove_extension(filename) + affix + OUT_EXT + suffix
            if system.non_empty_file(new_filename) and not FORCE:
                system.print_stderr("Error: file {nf} exists. Use FORCE to overwrite".
                                    format(nf=new_filename))
            else:
                system.write_file(new_filename, text)
                print(new_filename)

    return

#------------------------------------------------------------------------

if __name__ == '__main__':
    main()



#! /usr/bin/env python
#
# Test(s) for ../test_simple_main_example.py
#
# Notes:
# - For debugging the tested script, the ALLOW_SUBCOMMAND_TRACING environment
#   option shows tracing output normally suppressed by  unittest_wrapper.py.
# - This can be run as follows:
#   $ PYTHONPATH=".:$PYTHONPATH" python ./mezcla/tests/test_simple_main_example.py
#

"""Tests for simple_main_example module"""

# Standard packages
## TODO: from collections import defaultdict

# Installed packages
import pytest

# Local packages
from mezcla import debug
from mezcla.unittest_wrapper import TestWrapper
from mezcla.my_regex import my_re
from mezcla import system

# Note: Two references are used for the module to be tested:
#    THE_MODULE:	    global module object
#    TestIt.script_module:              path to file
import mezcla.simple_main_example as THE_MODULE

class TestSimpleMainExample(TestWrapper):
    """Class for testcase definition"""
    script_module = TestWrapper.get_testing_module_name(__file__, THE_MODULE)

    def test_data_file(self):
        """Tests run_script w/ data file"""
        debug.trace(4, f"TestIt.test_data_file(); self={self}")
        data = ["fubar1", "FUBAR2", "", "FuBART"]
        system.write_lines(self.temp_file, data)
        # Line mode
        output = self.run_script(options="", data_file=self.temp_file)
        self.do_assert(my_re.search(r"fubar1", output.strip()))
        self.do_assert(not my_re.search(r"FUBAR2", output.strip()))
        self.do_assert(not my_re.search(r"FuBART", output.strip()))
        # Para mode
        output = self.run_script(options="--para", data_file=self.temp_file)
        self.do_assert(my_re.search(r"FUBAR2", output.strip()))
        # Entire file
        output = self.run_script(options="--entire", data_file=self.temp_file)
        self.do_assert(my_re.search(r"FuBART", output.strip()))
        return

    def test_fitler_class(self):
        """Test SimpleFilter"""
        filter_inst = THE_MODULE.SimpleFilter(r"^\s*M(oo|u)law\s*$", flags=my_re.IGNORECASE)
        self.do_assert(filter_inst.include("MooLaw"))
        self.do_assert(filter_inst.include(" MULAW"))
        self.do_assert(not filter_inst.include("Mucho MooLaw"))
        self.do_assert(not filter_inst.include("MooLAW galore"))

if __name__ == '__main__':
    debug.trace_current_context()
    pytest.main([__file__])



#! /usr/bin/env python
#
# merge_notes.py: merge textual note files based on timestamps
#------------------------------------------------------------------------
# Sample input:
#
# - [some-notes.txt]
#   Thu 13 Nov 14
#
#      wrote great linear sort
#
#   Sat 15 Nov 14
#
#       rewrote not-so-great n log n sort
#
# - [more-notes.txt]
#   Fri 14 Nov 14
#
#      ran unit tests
#
#
# Sample output:
#
#   Thurs 13 Nov 14
#
#      wrote great linear sort
#
#   Fri 14 Nov 14
#
#      ran unit tests
#
#   Sat 15 Nov 14
#
#       rewrote not-so-great n log n sort
#
#-------------------------------------------------------------------------------
# Notes:
# - Includes indicator of original file and line number in output
#      [vm-plata-notes.txt:44]
#      Did this and that
#-------------------------------------------------------------------------------
# TODO:
# - *** Resolve problems with encoding (e.g., UTF-8): see ~/config/_master-note-info.list.08Jul20, such as following:
#      find: ‘/run/user/1000/gvfs’: Permission denied
#            ^ \342\200\230
# - ** Add option for maximum size (e.g., 1000 lines) to avoid inadvertantly incorporating large data files (e.g., sample-patient-notes.txt).
# - Allow for missing day-of-week and day (e.g., "Mar 16" => "Tues 1 Mar 16 [was Mar 16]").
#
#

"""Merge textual notes by dated entries"""

# Standard packages
import argparse
import datetime
import fileinput
import re
import sys
from collections import defaultdict

# Local packages
from mezcla import debug
from mezcla.my_regex import my_re
from mezcla import system

SKIP_BOM = system.getenv_bool("SKIP_BOM", False,
                              "Don't output byte order mark U+FEFF")

#...............................................................................

def resolve_date(textual_date, default_date=None):
    """Converts from textual DATE into datetime object (using optional DEFAULT value)"""
    # TODO: """... Supports following date formats: [WWW] dd MMM YY[YY]"
    # ALSO: clarify that default date used after parsing errors
    #
    # Note: Examples use first 5 of 8 of the datetime arguments:
    #     year, month, day, hour, minute
    # EX: resolve_date("1 Jan 00") => datetime.datetime(2000, 1, 1, 0, 0)
    # Note: following uses default date:
    # EX: resolve_date("0 Jan 00", datetime.datetime(2000, 1, 1, 0, 0)) => datetime.datetime(2000, 1, 1, 0, 0)
    # EX: resolve_date("Sun 18 Jul 2021") => datetime.datetime(2021, 7, 18, 0, 0)
    # Note: date component specifiers: %a: abbreviated weekday; %d day of month (2 digits); %b abbreviated month; %y year without century; %Y: year with century
    if default_date:
        debug.assertion(isinstance(default_date, datetime.datetime))
    date = default_date
    resolved = False
    for date_format in ["%a %d %b %y", "%d %b %y", "%a %d %b %Y", "%d %b %Y"]:
        try:
            date = datetime.datetime.strptime(textual_date, date_format)
            resolved = True
            break
        except ValueError:
            pass
    if not resolved:
        debug.trace_fmtd(2, "Warning: Unable to resolve date '{t}'", t=textual_date)
    debug.assertion(isinstance(date, datetime.datetime))
    debug.trace_fmtd(5, "resolve_date({t}, {d}) => {r}",
                     t=textual_date, d=default_date, r=date)
    return date


def main():
    """Entry point for script"""
    debug.trace(4, "main(): sys.argv=%s" % sys.argv)

    # Check command-line arguments
    parser = argparse.ArgumentParser(description="Merges ascii notes files")
    parser.add_argument("--ignore-dividers", default=False, action='store_true', help="Ignore lines consisting soley of dashes")
    parser.add_argument("--output-dividers", default=False, action='store_true', help="Output divider lines (80 dashes) betweens sections for different days")
    parser.add_argument("--show-file-info", default=False, action='store_true', help="Include filename and line number of original file in output")
    parser.add_argument("filename", nargs='+', default='-', help="Input filename")
    args = vars(parser.parse_args())
    debug.trace(5, "args = %s" % args)
    full_input_files = args['filename']
    ignore_dividers = args['ignore_dividers']
    output_dividers = args['output_dividers']
    show_file_info = args['show_file_info']

    # Initial defaults
    # note: initializes current date to dummy from way back when
    line_num = 0
    notes_hash = defaultdict(str)
    resolved_date = {}
    dummy_date = "1 Jan 1900"
    dummy_hour = "00:00:00"
    resolved_dummy_date = resolve_date(dummy_date)
    debug.assertion(resolved_dummy_date < resolve_date("1 Jan 00"))
    debug.assertion(dummy_hour in str(resolved_dummy_date))
    last_date = dummy_date
    last_resolved_date = resolved_dummy_date
    needs_source_info = True

    # Filter inaccessible files from input file list
    input_files = [f for f in full_input_files if system.file_exists(f)]
    inaccessible_files = system.difference(full_input_files, input_files)
    if inaccessible_files:
        system.print_stderr("Warning: ignoring {n} inaccessible files {fl}",
                            n=len(inaccessible_files), fl=inaccessible_files)
    
    # Read in all the notes line by line, saving text for notes entries keyed by date.
    # TODO: add in the text in groups of lines to allow for de-duplication across files (e.g., same entry from current note file and earlier version in restore directory)
    hook_utf8_replace = None
    if sys.version_info.major > 2:
        hook_utf8_replace = fileinput.hook_encoded('UTF-8', errors='replace')
    has_new_date = False
    for line in fileinput.input(input_files, openhook=hook_utf8_replace):
        line_num += 1
        # Note: strips leading and trailing spaces from line to facilitate regex
        # pattern matching, with raw line saved as original_line.
        original_line = line.strip("\n")
        line = line.strip()
        debug.trace(6, "L%d: %s" % (line_num, line))

        # Reset default date if first line in file
        if fileinput.isfirstline():
            debug.trace_fmtd(4, "new file: {f}", f=fileinput.filename())
            last_date = dummy_date
            last_resolved_date = resolved_dummy_date
            needs_source_info = True
            line_num = 1

        # Optionally ignore section dividers (20 or more dashes)
        if (ignore_dividers and re.search("^--------------------+$", line)):
            debug.trace_fmtd(5, "Ignoring divider at line {n}: {l}",
                             l=original_line, n=line_num)
            continue

        # Look for a new date in format Day dd Mon yy (e.g., "Fri 13 Nov 13")
        # Notes:
        # - Day and Mon are capitalized 3-letter abbreviations (i.e.., Sun, ..., Sat and Jan, ..., Dec)
        # - Source file and line information will be added for each new date
        # TODO: allow for a variety of date formats; allow for optional time
        new_date = last_date
        new_resolved_date = last_resolved_date
        # Ensure days of the week are abbreviated (with no more than 3 letters)
        line = re.sub(r"^(Sun|Mon|Tue|Wed|Thu|Fri|Sat)\w+day", r"\1", line, flags=re.IGNORECASE)
        line = re.sub(r"^(Tue)s (\d)", r"\1 \2", line, flags=re.IGNORECASE)
        line = re.sub(r"^(Thu)rs? (\d)", r"\1 \2", line, flags=re.IGNORECASE)
        # TODO: Ensure months are abbreviated
        ## EX: line = re.sub(r" (\d+) (Jan|Feb|Mar|Apr|May|Jun|Jul|Aug|Sep|Oct|Nov|Dec)\w* (\d+)", r" \1 \2 \3", line, flags=re.IGNORECASE)
        if (my_re.search(r"^([a-z][a-z][a-z] )?\d+ [a-z][a-z][a-z] \d+$", line, re.IGNORECASE)):
            new_date = my_re.group(0)
            needs_source_info = True
            has_new_date = True

            # Resolve date format, adding to hash if not already there
            # TODO: if not resolvable, report file and line number
            if new_date not in resolved_date:
                new_resolved_date = resolve_date(new_date, last_resolved_date)
                debug.assertion(dummy_hour in str(new_resolved_date))
                resolved_date[new_date] = new_resolved_date
                # TODO: only add source info if different date
                # needs_source_info = True

            # Update current date
            # note: used for subsequent lines without date specifications
            last_date = new_date
            last_resolved_date = new_resolved_date

            # Trace date resolution
            debug.trace_fmtd(5, "New date at line {n}: raw={raw}; resolved={new}\n", 
                             n=line_num, raw=new_date, new=new_resolved_date)
        else:
            debug.trace_fmt(7, "Ignoring non-date in date check at line {n}", n=line_num)
                            
        # Add optional source indicator to current date
        if show_file_info and needs_source_info:
            notes_hash[new_date] += "[src: {f}:{n}]\n".format(f=fileinput.filename(), 
                                                             n=fileinput.filelineno())
            needs_source_info = False
                
        # Add line to notes for current date
        # TODO: use resolved date as key so different specifications for same date output together without new date spec
        debug.assertion((not has_new_date) or (new_date != dummy_date))
        notes_hash[new_date] += original_line + "\n"
        has_new_date = False

    # Print the note entries sorted by resolved date.
    # Note:
    # - The sorting is based on the datetime.datetime type. If an error
    #   occurs, the problem might be due to the resolve_date function
    #   incorrectly returning a string instead of the proper datetime type.
    # - This is used for sake of debugging (e.g., tracing bad comparison input).
    def get_resolved_date(k):
        """Debugging accessor for resolved_date with tracing"""
        r = resolved_date.get(k, resolved_dummy_date)
        debug.trace_fmtd(7, "get_resolved_date({k}) => {r}", k=k, r=r)
        return r
    #
    debug.trace_fmtd(7, "notes_hash keys: {{\n{k}\n}}", 
                     k="\t\n".join([str(v) for v in notes_hash.keys()]))
    #
    if not SKIP_BOM:
        try:
            print("\uFEFF")
        except:
            system.print_exception_info("printing BOM")
    #
    for pos, date in enumerate(sorted(notes_hash.keys(), 
                                      key=get_resolved_date)):
        debug.trace_fmtd(6, "outputting notes for date {d} [resolved: {r}]", 
                         d=date, r=resolved_date.get(date))
        if output_dividers and (pos > 0):
            print("-" * 80)
        debug.trace_fmtd(6, "[src {f}:{n}]", skip_newline=True,
                         f=fileinput.filename(), n=fileinput.filelineno())
        try:
            print("%s\n\n" % notes_hash[date])
        except:
            system.print_exception_info("printing entry")

    return

#------------------------------------------------------------------------

if __name__ == '__main__':
    main()



## TPO: new version and docstring
"""
Package for Term Frequency (TF) Inversion Document Frequency (IDF): TF-IDF
"""
__version__ = "1.6"

## TODO
## # HACK: For relative imports to work in Python 3.6
## # See https://stackoverflow.com/questions/16981921/relative-imports-in-python-3.
## import os
## import sys;
## sys.path.append(os.path.dirname(os.path.abspath(__file__)))



#! /usr/bin/env python
#
# Test(s) for ../merge_files.py
#
# Notes:
# - For debugging the tested script, the ALLOW_SUBCOMMAND_TRACING environment
#   option shows tracing output normally suppressed by  unittest_wrapper.py.
# - This can be run as follows:
#   $ PYTHONPATH=".:$PYTHONPATH" python ./mezcla/tests/test_merge_files.py
#

"""Tests for merge_files module"""

# Standard packages
import re

# Installed packages
import pytest

# Local packages
from mezcla import debug
from mezcla import glue_helpers as gh
from mezcla import system
from mezcla.unittest_wrapper import TestWrapper

# Note: Two references are used for the module to be tested:
#    THE_MODULE:	    global module object
import mezcla.merge_files as THE_MODULE

class TestMergeFiles(TestWrapper):
    """Class for testcase definition"""
    script_file = TestWrapper.get_module_file_path(__file__)
    script_module = TestWrapper.get_testing_module_name(__file__)
    use_temp_base_dir = True            # treat TEMP_BASE as directory

    @pytest.mark.xfail
    def test_get_timestamp(self):
        """Ensure get_timestamp works as expected"""
        debug.trace(4, f"test_get_timestamp(); self={self}")
        lsb_date = gh.run('date +%F -r /proc/fb').strip().split('-')
        lts_year = lsb_date[0]
        lts_month = lsb_date[1]
        assert re.search(fr"({lts_year})-{lts_month}-\d\d \d\d:\d\d:\d\d", THE_MODULE.get_timestamp("/proc/fb"))

    def test_get_numeric_timestamp(self):
        """Ensure get_numeric_timestamp works as expected"""
        debug.trace(4, f"test_get_numeric_timestamp(); self={self}")
        # Note: /tmp should be newer than /etc
        first_timestamp = THE_MODULE.get_numeric_timestamp("/etc")
        second_timestamp = THE_MODULE.get_numeric_timestamp("/tmp")
        assert first_timestamp is not None
        assert second_timestamp is not None
        assert first_timestamp < second_timestamp

    @pytest.mark.xfail
    def test_simple_merge(self):
        """Make sure simple merge works"""
        debug.trace(4, "test_simple_merge()")
        tmp = self.temp_base
        system.write_lines(f"{tmp}/1.list", list(map(str, [0, 1, 2, 3])))
        gh.full_mkdir(f"{tmp}/backup")
        system.write_lines(f"{tmp}/backup/1.list", list(map(str, ['1', '2', '3'])))
        system.write_lines(f"{tmp}/other-1.list", list(map(str, ['1', '2', '3', '4'])))
        actual = self.run_script(
            env_options="IGNORE_TIMESTAMP=1",
            uses_stdin=True,
            data_file="",
            options=f"{tmp}/1.list {tmp}/other-1.list",
            )
        assert actual.split() == "0 1 2 3 4".split()

if __name__ == '__main__':
    debug.trace_current_context()
    pytest.main([__file__])
